#   Copyright 2012 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import strutils
from six.moves import http_client
import webob

from cinder.api import common
from cinder.api import extensions
from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import admin_actions
from cinder.api import validation
from cinder import backup
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder import rpc
from cinder import volume
from cinder.volume import volume_utils


LOG = logging.getLogger(__name__)


class AdminController(wsgi.Controller):
    """Abstract base class for AdminControllers."""

    collection = None  # api collection to extend

    # FIXME(clayg): this will be hard to keep up-to-date
    # Concrete classes can expand or over-ride

    def __init__(self, *args, **kwargs):
        super(AdminController, self).__init__(*args, **kwargs)
        # singular name of the resource
        self.resource_name = self.collection.rstrip('s')
        self.volume_api = volume.API()
        self.backup_api = backup.API()

    def _update(self, *args, **kwargs):
        raise NotImplementedError()

    def _get(self, *args, **kwargs):
        raise NotImplementedError()

    def _delete(self, *args, **kwargs):
        raise NotImplementedError()

    def validate_update(self, req, body):
        raise NotImplementedError()

    def _notify_reset_status(self, context, id, message):
        raise NotImplementedError()

    def authorize(self, context, action_name, target_obj=None):
        context.authorize(
            'volume_extension:%(resource)s_admin_actions:%(action)s' %
            {'resource': self.resource_name,
             'action': action_name}, target_obj=target_obj)

    def _remove_worker(self, context, id):
        # Remove the cleanup worker from the DB when we change a resource
        # status since it renders useless the entry.
        res = db.worker_destroy(context, resource_type=self.collection.title(),
                                resource_id=id)
        if res:
            LOG.debug('Worker entry for %s with id %s has been deleted.',
                      self.collection, id)

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.action('os-reset_status')
    def _reset_status(self, req, id, body):
        """Reset status on the resource."""

        def _clean_volume_attachment(context, id):
            attachments = (
                db.volume_attachment_get_all_by_volume_id(context, id))
            for attachment in attachments:
                db.volume_detached(context.elevated(), id, attachment.id)
            db.volume_admin_metadata_delete(context.elevated(), id,
                                            'attached_mode')

        context = req.environ['cinder.context']
        update = self.validate_update(req, body=body)
        msg = "Updating %(resource)s '%(id)s' with '%(update)r'"
        LOG.debug(msg, {'resource': self.resource_name, 'id': id,
                        'update': update})

        # calling notifier here for volumeStatusUpdate is deprecated.
        # Will be replaced with _notify_reset_status()
        notifier_info = dict(id=id, update=update)
        notifier = rpc.get_notifier('volumeStatusUpdate')
        notifier.info(context, self.collection + '.reset_status.start',
                      notifier_info)
        self._notify_reset_status(context, id, 'reset_status.start')

        # Not found exception will be handled at the wsgi level
        self._update(context, id, update)
        self._remove_worker(context, id)
        if update.get('attach_status') == 'detached':
            _clean_volume_attachment(context, id)

        notifier.info(context, self.collection + '.reset_status.end',
                      notifier_info)
        self._notify_reset_status(context, id, 'reset_status.end')

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.action('os-force_delete')
    def _force_delete(self, req, id, body):
        """Delete a resource, bypassing the check that it must be available."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        resource = self._get(context, id)
        self.authorize(context, 'force_delete', target_obj=resource)
        self._delete(context, resource, force=True)


class VolumeAdminController(AdminController):
    """AdminController for Volumes."""

    collection = 'volumes'

    def _notify_reset_status(self, context, id, message):
        volume = objects.Volume.get_by_id(context, id)
        volume_utils.notify_about_volume_usage(context, volume,
                                               message)

    def _update(self, *args, **kwargs):
        context = args[0]
        volume_id = args[1]
        volume = objects.Volume.get_by_id(context, volume_id)
        self.authorize(context, 'reset_status', target_obj=volume)
        db.volume_update(*args, **kwargs)

    def _get(self, *args, **kwargs):
        return self.volume_api.get(*args, **kwargs)

    def _delete(self, *args, **kwargs):
        return self.volume_api.delete(*args, **kwargs)

    @validation.schema(admin_actions.reset)
    def validate_update(self, req, body):
        update = {}
        body = body['os-reset_status']
        status = body.get('status', None)
        attach_status = body.get('attach_status', None)
        migration_status = body.get('migration_status', None)

        if status:
            update['status'] = status.lower()

        if attach_status:
            update['attach_status'] = attach_status.lower()

        if migration_status:
            update['migration_status'] = migration_status.lower()
            if update['migration_status'] == 'none':
                update['migration_status'] = None

        return update

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.action('os-force_detach')
    @validation.schema(admin_actions.force_detach)
    def _force_detach(self, req, id, body):
        """Roll back a bad detach after the volume been disconnected."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self._get(context, id)
        self.authorize(context, 'force_detach', target_obj=volume)
        connector = body['os-force_detach'].get('connector', None)

        try:
            self.volume_api.terminate_connection(context, volume, connector)
        except exception.VolumeBackendAPIException as error:
            msg = _("Unable to terminate volume connection from backend.")
            raise webob.exc.HTTPInternalServerError(explanation=msg)

        attachment_id = body['os-force_detach'].get('attachment_id', None)

        try:
            self.volume_api.detach(context, volume, attachment_id)
        except messaging.RemoteError as error:
            if error.exc_type in ['VolumeAttachmentNotFound',
                                  'InvalidVolume']:
                msg = _("Error force detaching volume - %(err_type)s: "
                        "%(err_msg)s") % {'err_type': error.exc_type,
                                          'err_msg': error.value}
                raise webob.exc.HTTPBadRequest(explanation=msg)
            else:
                # There are also few cases where force-detach call could fail
                # due to db or volume driver errors. These errors shouldn't
                # be exposed to the user and in such cases it should raise
                # 500 error.
                raise

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.action('os-migrate_volume')
    @validation.schema(admin_actions.migrate_volume, mv.V2_BASE_VERSION,
                       mv.get_prior_version(mv.VOLUME_MIGRATE_CLUSTER))
    @validation.schema(admin_actions.migrate_volume_v316,
                       mv.VOLUME_MIGRATE_CLUSTER)
    def _migrate_volume(self, req, id, body):
        """Migrate a volume to the specified host."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self._get(context, id)
        self.authorize(context, 'migrate_volume', target_obj=volume)
        params = body['os-migrate_volume']

        cluster_name, host = common.get_cluster_host(req, params,
                                                     mv.VOLUME_MIGRATE_CLUSTER)
        force_host_copy = strutils.bool_from_string(params.get(
            'force_host_copy', False), strict=True)
        lock_volume = strutils.bool_from_string(params.get(
            'lock_volume', False), strict=True)
        self.volume_api.migrate_volume(context, volume, host, cluster_name,
                                       force_host_copy, lock_volume)

    @wsgi.action('os-migrate_volume_completion')
    @validation.schema(admin_actions.migrate_volume_completion)
    def _migrate_volume_completion(self, req, id, body):
        """Complete an in-progress migration."""
        context = req.environ['cinder.context']
        # Not found exception will be handled at the wsgi level
        volume = self._get(context, id)
        self.authorize(context, 'migrate_volume_completion', target_obj=volume)
        params = body['os-migrate_volume_completion']
        new_volume_id = params['new_volume']
        # Not found exception will be handled at the wsgi level
        new_volume = self._get(context, new_volume_id)
        error = params.get('error', False)
        ret = self.volume_api.migrate_volume_completion(context, volume,
                                                        new_volume, error)
        return {'save_volume_id': ret}


class SnapshotAdminController(AdminController):
    """AdminController for Snapshots."""

    collection = 'snapshots'

    def _notify_reset_status(self, context, id, message):
        snapshot = objects.Snapshot.get_by_id(context, id)
        volume_utils.notify_about_snapshot_usage(context, snapshot,
                                                 message)

    @validation.schema(admin_actions.reset_status_snapshot)
    def validate_update(self, req, body):
        status = body['os-reset_status']['status']
        update = {'status': status.lower()}
        return update

    def _update(self, *args, **kwargs):
        context = args[0]
        snapshot_id = args[1]
        fields = args[2]
        snapshot = objects.Snapshot.get_by_id(context, snapshot_id)
        self.authorize(context, 'reset_status', target_obj=snapshot)
        snapshot.update(fields)
        snapshot.save()

    def _get(self, *args, **kwargs):
        return self.volume_api.get_snapshot(*args, **kwargs)

    def _delete(self, *args, **kwargs):
        return self.volume_api.delete_snapshot(*args, **kwargs)


class BackupAdminController(AdminController):
    """AdminController for Backups."""

    collection = 'backups'

    def _notify_reset_status(self, context, id, message):
        backup = objects.Backup.get_by_id(context, id)
        volume_utils.notify_about_backup_usage(context, backup,
                                               message)

    def _get(self, *args, **kwargs):
        return self.backup_api.get(*args, **kwargs)

    def _delete(self, *args, **kwargs):
        return self.backup_api.delete(*args, **kwargs)

    @wsgi.response(http_client.ACCEPTED)
    @wsgi.action('os-reset_status')
    @validation.schema(admin_actions.reset_status_backup)
    def _reset_status(self, req, id, body):
        """Reset status on the resource."""
        context = req.environ['cinder.context']
        status = body['os-reset_status']['status']
        update = {'status': status.lower()}
        msg = "Updating %(resource)s '%(id)s' with '%(update)r'"
        LOG.debug(msg, {'resource': self.resource_name, 'id': id,
                        'update': update})

        notifier_info = {'id': id, 'update': update}
        notifier = rpc.get_notifier('backupStatusUpdate')
        notifier.info(context, self.collection + '.reset_status.start',
                      notifier_info)
        self._notify_reset_status(context, id, 'reset_status.start')

        # Not found exception will be handled at the wsgi level
        self.backup_api.reset_status(context=context, backup_id=id,
                                     status=update['status'])


class Admin_actions(extensions.ExtensionDescriptor):
    """Enable admin actions."""

    name = "AdminActions"
    alias = "os-admin-actions"
    updated = "2012-08-25T00:00:00+00:00"

    def get_controller_extensions(self):
        exts = []
        for class_ in (VolumeAdminController, SnapshotAdminController,
                       BackupAdminController):
            controller = class_()
            extension = extensions.ControllerExtension(
                self, class_.collection, controller)
            exts.append(extension)
        return exts



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder import context
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_volume
from cinder.tests.unit.volume.drivers.dell_emc import vxflexos


class TestAttachDetachVolume(vxflexos.TestVxFlexOSDriver):

    def setUp(self):
        super(TestAttachDetachVolume, self).setUp()
        ctx = context.RequestContext('fake', 'fake', auth_token=True)
        self.fake_path = '/fake/path/vol-xx'
        self.volume = fake_volume.fake_volume_obj(
            ctx, **{'provider_id': fake.PROVIDER_ID})
        self.driver.connector = FakeConnector()

    def test_attach_volume(self):
        path = self.driver._sio_attach_volume(self.volume)
        self.assertEqual(self.fake_path, path)

    def test_detach_volume(self):
        self.driver._sio_detach_volume(self.volume)


class FakeConnector(object):
    def connect_volume(self, connection_properties):
        return {'path': '/fake/path/vol-xx'}

    def disconnect_volume(self, connection_properties, volume):
        return None



# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For HostManager
"""

from datetime import datetime
from datetime import timedelta
import ddt

import mock
from oslo_serialization import jsonutils
from oslo_utils import timeutils

from cinder.common import constants
from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.scheduler import filters
from cinder.scheduler import host_manager
from cinder import test
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit.objects import test_service


class FakeFilterClass1(filters.BaseBackendFilter):
    def backend_passes(self, host_state, filter_properties):
        pass


class FakeFilterClass2(filters.BaseBackendFilter):
    def backend_passes(self, host_state, filter_properties):
        pass


class FakeFilterClass3(filters.BaseHostFilter):
    def host_passes(self, host_state, filter_properties):
        return host_state.get('volume_backend_name') == \
            filter_properties.get('volume_type')['volume_backend_name']


@ddt.ddt
class HostManagerTestCase(test.TestCase):
    """Test case for HostManager class."""

    def setUp(self):
        super(HostManagerTestCase, self).setUp()
        self.host_manager = host_manager.HostManager()
        self.fake_backends = [host_manager.BackendState('fake_be%s' % x, None)
                              for x in range(1, 5)]
        # For a second scheduler service.
        self.host_manager_1 = host_manager.HostManager()

    @mock.patch(
        'cinder.scheduler.filters.BackendFilterHandler.get_all_classes')
    def test_initialzie_with_default_filters(self, mock_get_all_classes):
        self.flags(scheduler_default_filters=
                   'FakeFilterClass1,FakeFilterClass2')
        mock_get_all_classes.return_value = [
            FakeFilterClass1, FakeFilterClass2, FakeFilterClass3]
        h_manager = host_manager.HostManager()
        self.assertListEqual([FakeFilterClass1, FakeFilterClass2],
                             h_manager.enabled_filters)

    def test_choose_backend_filters_not_found(self):
        self.host_manager.filter_classes = [FakeFilterClass1,
                                            FakeFilterClass2]
        self.assertRaises(exception.SchedulerHostFilterNotFound,
                          self.host_manager._choose_backend_filters,
                          'FakeFilterClass3')

    def test_choose_backend_filters(self):
        self.host_manager.filter_classes = [FakeFilterClass1,
                                            FakeFilterClass2]

        # Test 'volume' returns 1 correct function
        filter_classes = self.host_manager._choose_backend_filters(
            'FakeFilterClass2')
        self.assertEqual(1, len(filter_classes))
        self.assertEqual('FakeFilterClass2', filter_classes[0].__name__)

    @mock.patch('cinder.scheduler.host_manager.HostManager.'
                '_choose_backend_filters')
    def test_get_filtered_backends(self, _mock_choose_backend_filters):
        filter_class = FakeFilterClass1
        mock_func = mock.Mock()
        mock_func.return_value = True
        filter_class._filter_one = mock_func
        _mock_choose_backend_filters.return_value = [filter_class]

        fake_properties = {'moo': 1, 'cow': 2}
        expected = []
        for fake_backend in self.fake_backends:
            expected.append(mock.call(fake_backend, fake_properties))
        host_manager1 = host_manager.HostManager()
        result = host_manager1.get_filtered_backends(self.fake_backends,
                                                     fake_properties)
        self.assertEqual(expected, mock_func.call_args_list)
        self.assertEqual(set(self.fake_backends), set(result))

    @mock.patch(
        'cinder.scheduler.host_manager.HostManager._is_just_initialized')
    @mock.patch('cinder.scheduler.host_manager.HostManager._get_updated_pools')
    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_update_service_capabilities(self, _mock_utcnow,
                                         _mock_get_updated_pools,
                                         _mock_is_just_initialized):
        service_states = self.host_manager.service_states
        self.assertDictEqual({}, service_states)
        _mock_is_just_initialized.return_value = False
        _mock_utcnow.side_effect = [31338, 31339]

        _mock_get_updated_pools.return_value = []
        timestamp = jsonutils.to_primitive(datetime.utcnow())
        host1_volume_capabs = dict(free_capacity_gb=4321, timestamp=timestamp)
        host1_old_volume_capabs = dict(free_capacity_gb=1, timestamp=timestamp)
        host2_volume_capabs = dict(free_capacity_gb=5432)
        host3_volume_capabs = dict(free_capacity_gb=6543)

        service_name = 'volume'
        # The host manager receives a deserialized timestamp
        timestamp = datetime.strptime(timestamp, timeutils.PERFECT_TIME_FORMAT)
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      host1_volume_capabs,
                                                      None, timestamp)
        # It'll ignore older updates
        old_timestamp = timestamp - timedelta(hours=1)
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      host1_old_volume_capabs,
                                                      None, old_timestamp)
        self.host_manager.update_service_capabilities(service_name, 'host2',
                                                      host2_volume_capabs,
                                                      None, None)
        self.host_manager.update_service_capabilities(service_name, 'host3',
                                                      host3_volume_capabs,
                                                      None, None)

        # Make sure dictionary isn't re-assigned
        self.assertEqual(service_states, self.host_manager.service_states)

        host1_volume_capabs['timestamp'] = timestamp
        host2_volume_capabs['timestamp'] = 31338
        host3_volume_capabs['timestamp'] = 31339

        expected = {'host1': host1_volume_capabs,
                    'host2': host2_volume_capabs,
                    'host3': host3_volume_capabs}
        self.assertDictEqual(expected, service_states)

    @mock.patch(
        'cinder.scheduler.host_manager.HostManager._is_just_initialized')
    @mock.patch(
        'cinder.scheduler.host_manager.HostManager.get_usage_and_notify')
    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_update_and_notify_service_capabilities_case1(
            self, _mock_utcnow,
            _mock_get_usage_and_notify,
            _mock_is_just_initialized):

        _mock_utcnow.side_effect = [31337, 31338, 31339]
        _mock_is_just_initialized.return_value = False
        service_name = 'volume'

        capab1 = {'pools': [{
                  'pool_name': 'pool1', 'thick_provisioning_support': True,
                  'thin_provisioning_support': False, 'total_capacity_gb': 10,
                  'free_capacity_gb': 10, 'max_over_subscription_ratio': '1',
                  'provisioned_capacity_gb': 0, 'allocated_capacity_gb': 0,
                  'reserved_percentage': 0}]}

        # Run 1:
        # capa: capa1
        # S0: update_service_capabilities()
        # S0: notify_service_capabilities()
        # S1: update_service_capabilities()
        #
        # notify capab1 to ceilometer by S0
        #

        # S0: update_service_capabilities()
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      capab1, None, None)
        self.assertDictEqual(dict(dict(timestamp=31337), **capab1),
                             self.host_manager.service_states['host1'])

        # S0: notify_service_capabilities()
        self.host_manager.notify_service_capabilities(service_name, 'host1',
                                                      capab1, None)
        self.assertDictEqual(dict(dict(timestamp=31337), **capab1),
                             self.host_manager.service_states['host1'])
        self.assertDictEqual(
            dict(dict(timestamp=31338), **capab1),
            self.host_manager.service_states_last_update['host1'])

        # notify capab1 to ceilometer by S0
        self.assertEqual(1, _mock_get_usage_and_notify.call_count)

        # S1: update_service_capabilities()
        self.host_manager_1.update_service_capabilities(service_name, 'host1',
                                                        capab1, None, None)

        self.assertDictEqual(dict(dict(timestamp=31339), **capab1),
                             self.host_manager_1.service_states['host1'])

    @mock.patch(
        'cinder.scheduler.host_manager.HostManager._is_just_initialized')
    @mock.patch(
        'cinder.scheduler.host_manager.HostManager.get_usage_and_notify')
    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_update_and_notify_service_capabilities_case2(
            self, _mock_utcnow,
            _mock_get_usage_and_notify,
            _mock_is_just_initialized):

        _mock_utcnow.side_effect = [31340, 31341, 31342]
        _mock_is_just_initialized.return_value = False

        service_name = 'volume'

        capab1 = {'pools': [{
                  'pool_name': 'pool1', 'thick_provisioning_support': True,
                  'thin_provisioning_support': False, 'total_capacity_gb': 10,
                  'free_capacity_gb': 10, 'max_over_subscription_ratio': '1',
                  'provisioned_capacity_gb': 0, 'allocated_capacity_gb': 0,
                  'reserved_percentage': 0}]}

        self.host_manager.service_states['host1'] = (
            dict(dict(timestamp=31337), **capab1))
        self.host_manager.service_states_last_update['host1'] = (
            dict(dict(timestamp=31338), **capab1))
        self.host_manager_1.service_states['host1'] = (
            dict(dict(timestamp=31339), **capab1))

        # Run 2:
        # capa: capa1
        # S0: update_service_capabilities()
        # S1: update_service_capabilities()
        # S1: notify_service_capabilities()
        #
        # Don't notify capab1 to ceilometer.

        # S0: update_service_capabilities()
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      capab1, None, None)

        self.assertDictEqual(dict(dict(timestamp=31340), **capab1),
                             self.host_manager.service_states['host1'])

        self.assertDictEqual(
            dict(dict(timestamp=31338), **capab1),
            self.host_manager.service_states_last_update['host1'])

        # S1: update_service_capabilities()
        self.host_manager_1.update_service_capabilities(service_name, 'host1',
                                                        capab1, None, None)

        self.assertDictEqual(dict(dict(timestamp=31341), **capab1),
                             self.host_manager_1.service_states['host1'])

        self.assertDictEqual(
            dict(dict(timestamp=31339), **capab1),
            self.host_manager_1.service_states_last_update['host1'])

        # S1: notify_service_capabilities()
        self.host_manager_1.notify_service_capabilities(service_name, 'host1',
                                                        capab1, None)

        self.assertDictEqual(dict(dict(timestamp=31341), **capab1),
                             self.host_manager_1.service_states['host1'])

        self.assertDictEqual(
            self.host_manager_1.service_states_last_update['host1'],
            dict(dict(timestamp=31339), **capab1))

        # Don't notify capab1 to ceilometer.
        self.assertEqual(0, _mock_get_usage_and_notify.call_count)

    @mock.patch(
        'cinder.scheduler.host_manager.HostManager.get_usage_and_notify')
    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_update_and_notify_service_capabilities_case3(
            self, _mock_utcnow,
            _mock_get_usage_and_notify):

        _mock_utcnow.side_effect = [31343, 31344, 31345]

        service_name = 'volume'

        capab1 = {'pools': [{
                  'pool_name': 'pool1', 'thick_provisioning_support': True,
                  'thin_provisioning_support': False, 'total_capacity_gb': 10,
                  'free_capacity_gb': 10, 'max_over_subscription_ratio': '1',
                  'provisioned_capacity_gb': 0, 'allocated_capacity_gb': 0,
                  'reserved_percentage': 0}]}

        self.host_manager.service_states['host1'] = (
            dict(dict(timestamp=31340), **capab1))
        self.host_manager.service_states_last_update['host1'] = (
            dict(dict(timestamp=31338), **capab1))
        self.host_manager_1.service_states['host1'] = (
            dict(dict(timestamp=31341), **capab1))
        self.host_manager_1.service_states_last_update['host1'] = (
            dict(dict(timestamp=31339), **capab1))

        # Run 3:
        # capa: capab1
        # S0: notify_service_capabilities()
        # S0: update_service_capabilities()
        # S1: update_service_capabilities()
        #
        # Don't notify capab1 to ceilometer.

        # S0: notify_service_capabilities()
        self.host_manager.notify_service_capabilities(service_name, 'host1',
                                                      capab1, None)
        self.assertDictEqual(
            dict(dict(timestamp=31338), **capab1),
            self.host_manager.service_states_last_update['host1'])

        self.assertDictEqual(dict(dict(timestamp=31340), **capab1),
                             self.host_manager.service_states['host1'])

        # Don't notify capab1 to ceilometer.
        self.assertEqual(0, _mock_get_usage_and_notify.call_count)

        # S0: update_service_capabilities()
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      capab1, None, None)

        self.assertDictEqual(
            dict(dict(timestamp=31340), **capab1),
            self.host_manager.service_states_last_update['host1'])

        self.assertDictEqual(dict(dict(timestamp=31344), **capab1),
                             self.host_manager.service_states['host1'])

        # S1: update_service_capabilities()
        self.host_manager_1.update_service_capabilities(service_name, 'host1',
                                                        capab1, None, None)
        self.assertDictEqual(dict(dict(timestamp=31345), **capab1),
                             self.host_manager_1.service_states['host1'])

        self.assertDictEqual(
            dict(dict(timestamp=31341), **capab1),
            self.host_manager_1.service_states_last_update['host1'])

    @mock.patch(
        'cinder.scheduler.host_manager.HostManager.get_usage_and_notify')
    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_update_and_notify_service_capabilities_case4(
            self, _mock_utcnow,
            _mock_get_usage_and_notify):

        _mock_utcnow.side_effect = [31346, 31347, 31348]

        service_name = 'volume'

        capab1 = {'pools': [{
                  'pool_name': 'pool1', 'thick_provisioning_support': True,
                  'thin_provisioning_support': False, 'total_capacity_gb': 10,
                  'free_capacity_gb': 10, 'max_over_subscription_ratio': '1',
                  'provisioned_capacity_gb': 0, 'allocated_capacity_gb': 0,
                  'reserved_percentage': 0}]}

        self.host_manager.service_states['host1'] = (
            dict(dict(timestamp=31344), **capab1))
        self.host_manager.service_states_last_update['host1'] = (
            dict(dict(timestamp=31340), **capab1))
        self.host_manager_1.service_states['host1'] = (
            dict(dict(timestamp=31345), **capab1))
        self.host_manager_1.service_states_last_update['host1'] = (
            dict(dict(timestamp=31341), **capab1))

        capab2 = {'pools': [{
                  'pool_name': 'pool1', 'thick_provisioning_support': True,
                  'thin_provisioning_support': False, 'total_capacity_gb': 10,
                  'free_capacity_gb': 9, 'max_over_subscription_ratio': '1',
                  'provisioned_capacity_gb': 1, 'allocated_capacity_gb': 1,
                  'reserved_percentage': 0}]}

        # Run 4:
        # capa: capab2
        # S0: update_service_capabilities()
        # S1: notify_service_capabilities()
        # S1: update_service_capabilities()
        #
        # notify capab2 to ceilometer.

        # S0: update_service_capabilities()
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      capab2, None, None)
        self.assertDictEqual(
            dict(dict(timestamp=31340), **capab1),
            self.host_manager.service_states_last_update['host1'])

        self.assertDictEqual(dict(dict(timestamp=31346), **capab2),
                             self.host_manager.service_states['host1'])

        # S1: notify_service_capabilities()
        self.host_manager_1.notify_service_capabilities(service_name, 'host1',
                                                        capab2, None)
        self.assertDictEqual(dict(dict(timestamp=31345), **capab1),
                             self.host_manager_1.service_states['host1'])

        self.assertDictEqual(
            dict(dict(timestamp=31347), **capab2),
            self.host_manager_1.service_states_last_update['host1'])

        # notify capab2 to ceilometer.
        self.assertLess(0, _mock_get_usage_and_notify.call_count)

        # S1: update_service_capabilities()
        self.host_manager_1.update_service_capabilities(service_name, 'host1',
                                                        capab2, None, None)
        self.assertDictEqual(dict(dict(timestamp=31348), **capab2),
                             self.host_manager_1.service_states['host1'])

        self.assertDictEqual(
            dict(dict(timestamp=31347), **capab2),
            self.host_manager_1.service_states_last_update['host1'])

    @mock.patch(
        'cinder.scheduler.host_manager.HostManager.get_usage_and_notify')
    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_update_and_notify_service_capabilities_case5(
            self, _mock_utcnow,
            _mock_get_usage_and_notify):

        _mock_utcnow.side_effect = [31349, 31350, 31351]

        service_name = 'volume'

        capab1 = {'pools': [{
                  'pool_name': 'pool1', 'thick_provisioning_support': True,
                  'thin_provisioning_support': False, 'total_capacity_gb': 10,
                  'free_capacity_gb': 10, 'max_over_subscription_ratio': '1',
                  'provisioned_capacity_gb': 0, 'allocated_capacity_gb': 0,
                  'reserved_percentage': 0}]}

        capab2 = {'pools': [{
                  'pool_name': 'pool1', 'thick_provisioning_support': True,
                  'thin_provisioning_support': False, 'total_capacity_gb': 10,
                  'free_capacity_gb': 9, 'max_over_subscription_ratio': '1',
                  'provisioned_capacity_gb': 1, 'allocated_capacity_gb': 1,
                  'reserved_percentage': 0}]}

        self.host_manager.service_states['host1'] = (
            dict(dict(timestamp=31346), **capab2))
        self.host_manager.service_states_last_update['host1'] = (
            dict(dict(timestamp=31340), **capab1))
        self.host_manager_1.service_states['host1'] = (
            dict(dict(timestamp=31348), **capab2))
        self.host_manager_1.service_states_last_update['host1'] = (
            dict(dict(timestamp=31347), **capab2))

        # Run 5:
        # capa: capa2
        # S0: notify_service_capabilities()
        # S0: update_service_capabilities()
        # S1: update_service_capabilities()
        #
        # This is the special case not handled.
        # 1) capab is changed (from capab1 to capab2)
        # 2) S1 has already notify the capab2 in Run 4.
        # 3) S0 just got update_service_capabilities() in Run 4.
        # 4) S0 got notify_service_capabilities() immediately in next run,
        #    here is Run 5.
        #    S0 has no ways to know whether other scheduler (here is S1) who
        #    has noitified the changed capab2 or not. S0 just thinks it's his
        #    own turn to notify the changed capab2.
        #    In this case, we have notified the same capabilities twice.
        #
        # S0: notify_service_capabilities()
        self.host_manager.notify_service_capabilities(service_name, 'host1',
                                                      capab2, None)
        self.assertDictEqual(
            dict(dict(timestamp=31349), **capab2),
            self.host_manager.service_states_last_update['host1'])

        self.assertDictEqual(dict(dict(timestamp=31346), **capab2),
                             self.host_manager.service_states['host1'])

        # S0 notify capab2 to ceilometer.
        self.assertLess(0, _mock_get_usage_and_notify.call_count)

        # S0: update_service_capabilities()
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      capab2, None, None)
        self.assertDictEqual(
            dict(dict(timestamp=31349), **capab2),
            self.host_manager.service_states_last_update['host1'])

        self.assertDictEqual(dict(dict(timestamp=31350), **capab2),
                             self.host_manager.service_states['host1'])

        # S1: update_service_capabilities()
        self.host_manager_1.update_service_capabilities(service_name, 'host1',
                                                        capab2, None, None)

        self.assertDictEqual(
            dict(dict(timestamp=31348), **capab2),
            self.host_manager_1.service_states_last_update['host1'])

        self.assertDictEqual(dict(dict(timestamp=31351), **capab2),
                             self.host_manager_1.service_states['host1'])

    @mock.patch('cinder.objects.service.Service.is_up',
                new_callable=mock.PropertyMock)
    @mock.patch('cinder.db.service_get_all')
    def test_has_all_capabilities(self, _mock_service_get_all,
                                  _mock_service_is_up):
        _mock_service_is_up.return_value = True
        services = [
            dict(id=1, host='host1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='a3a593da-7f8d-4bb7-8b4c-f2bc1e0b4824'),
            dict(id=2, host='host2', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='4200b32b-0bf9-436c-86b2-0675f6ac218e'),
            dict(id=3, host='host3', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='6d91e7f5-ca17-4e3b-bf4f-19ca77166dd7'),
        ]
        _mock_service_get_all.return_value = services
        # Create host_manager again to let db.service_get_all mock run
        self.host_manager = host_manager.HostManager()
        self.assertFalse(self.host_manager.has_all_capabilities())

        timestamp = jsonutils.to_primitive(datetime.utcnow())
        host1_volume_capabs = dict(free_capacity_gb=4321)
        host2_volume_capabs = dict(free_capacity_gb=5432)
        host3_volume_capabs = dict(free_capacity_gb=6543)

        service_name = 'volume'
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      host1_volume_capabs,
                                                      None, timestamp)
        self.assertFalse(self.host_manager.has_all_capabilities())
        self.host_manager.update_service_capabilities(service_name, 'host2',
                                                      host2_volume_capabs,
                                                      None, timestamp)
        self.assertFalse(self.host_manager.has_all_capabilities())
        self.host_manager.update_service_capabilities(service_name, 'host3',
                                                      host3_volume_capabs,
                                                      None, timestamp)
        self.assertTrue(self.host_manager.has_all_capabilities())

    @mock.patch('cinder.objects.service.Service.is_up',
                new_callable=mock.PropertyMock)
    @mock.patch('cinder.db.service_get_all')
    def test_first_receive_capabilities_case1(self, _mock_service_get_all,
                                              _mock_service_is_up):
        # No volume service startup
        self.assertFalse(self.host_manager.first_receive_capabilities())
        services = [
            dict(id=1, host='host1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='06acda71-b3b4-4f1b-8d87-db5c47e7ebd2', )
        ]
        _mock_service_get_all.return_value = services
        _mock_service_is_up.return_value = True

        timestamp = jsonutils.to_primitive(datetime.utcnow())
        host1_volume_capabs = dict(free_capacity_gb=4321)

        service_name = 'volume'
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      host1_volume_capabs,
                                                      None, timestamp)
        self.assertTrue(self.host_manager.first_receive_capabilities())

    @mock.patch('cinder.objects.service.Service.is_up',
                new_callable=mock.PropertyMock)
    @mock.patch('cinder.db.service_get_all')
    def test_first_receive_capabilities_case2(self, _mock_service_get_all,
                                              _mock_service_is_up):
        _mock_service_is_up.return_value = True
        services = [
            dict(id=1, host='host1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='36ede0e2-1b3c-41b0-9cd3-66e1f56dc959'),
            dict(id=2, host='host2', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='b124e8dc-bf5f-4923-802d-27153ac7fe56'),
            dict(id=3, host='host3', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='4d0b1c5e-ce3c-424e-b2f4-a09a0f54d328'),
        ]
        _mock_service_get_all.return_value = services
        # Create host_manager again to let db.service_get_all mock run
        self.host_manager = host_manager.HostManager()
        self.assertFalse(self.host_manager.first_receive_capabilities())

        timestamp = jsonutils.to_primitive(datetime.utcnow())
        host1_volume_capabs = dict(free_capacity_gb=4321)
        host2_volume_capabs = dict(free_capacity_gb=5432)
        host3_volume_capabs = dict(free_capacity_gb=6543)

        service_name = 'volume'
        self.host_manager.update_service_capabilities(service_name, 'host1',
                                                      host1_volume_capabs,
                                                      None, timestamp)
        self.assertFalse(self.host_manager.first_receive_capabilities())
        self.host_manager.update_service_capabilities(service_name, 'host2',
                                                      host2_volume_capabs,
                                                      None, timestamp)
        self.assertFalse(self.host_manager.first_receive_capabilities())
        self.host_manager.update_service_capabilities(service_name, 'host3',
                                                      host3_volume_capabs,
                                                      None, timestamp)
        self.assertTrue(self.host_manager.first_receive_capabilities())

    @mock.patch('cinder.db.service_get_all')
    @mock.patch('cinder.objects.service.Service.is_up',
                new_callable=mock.PropertyMock)
    @mock.patch('oslo_utils.timeutils.utcnow')
    def test_update_and_get_pools(self, _mock_utcnow,
                                  _mock_service_is_up,
                                  _mock_service_get_all):
        """Test interaction between update and get_pools

        This test verifies that each time that get_pools is called it gets the
        latest copy of service_capabilities, which is timestamped with the
        current date/time.
        """
        context = 'fake_context'
        dates = [datetime.fromtimestamp(400), datetime.fromtimestamp(401),
                 datetime.fromtimestamp(402)]
        _mock_utcnow.side_effect = dates

        services = [
            # This is the first call to utcnow()
            dict(id=1, host='host1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='6d91e7f5-ca17-4e3b-bf4f-19ca77166dd7',)
        ]

        mocked_service_states = {
            'host1': dict(volume_backend_name='AAA',
                          total_capacity_gb=512, free_capacity_gb=200,
                          timestamp=dates[1], reserved_percentage=0),
        }

        _mock_service_get_all.return_value = services
        _mock_service_is_up.return_value = True
        _mock_warning = mock.Mock()
        host_manager.LOG.warn = _mock_warning

        host_volume_capabs = dict(free_capacity_gb=4321)

        service_name = 'volume'
        with mock.patch.dict(self.host_manager.service_states,
                             mocked_service_states):
            self.host_manager.update_service_capabilities(service_name,
                                                          'host1',
                                                          host_volume_capabs,
                                                          None, None)
            res = self.host_manager.get_pools(context)
            self.assertEqual(1, len(res))
            self.assertEqual(dates[1], res[0]['capabilities']['timestamp'])

    @mock.patch('cinder.objects.Service.is_up', True)
    def test_get_all_backend_states_cluster(self):
        """Test get_all_backend_states when we have clustered services.

        Confirm that clustered services are grouped and that only the latest
        of the capability reports is relevant.
        """
        ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)

        cluster_name = 'cluster'
        db.cluster_create(ctxt, {'name': cluster_name,
                                 'binary': constants.VOLUME_BINARY})

        services = (
            db.service_create(ctxt,
                              {'host': 'clustered_host_1',
                               'topic': constants.VOLUME_TOPIC,
                               'binary': constants.VOLUME_BINARY,
                               'cluster_name': cluster_name,
                               'created_at': timeutils.utcnow()}),
            # Even if this service is disabled, since it belongs to an enabled
            # cluster, it's not really disabled.
            db.service_create(ctxt,
                              {'host': 'clustered_host_2',
                               'topic': constants.VOLUME_TOPIC,
                               'binary': constants.VOLUME_BINARY,
                               'disabled': True,
                               'cluster_name': cluster_name,
                               'created_at': timeutils.utcnow()}),
            db.service_create(ctxt,
                              {'host': 'clustered_host_3',
                               'topic': constants.VOLUME_TOPIC,
                               'binary': constants.VOLUME_BINARY,
                               'cluster_name': cluster_name,
                               'created_at': timeutils.utcnow()}),
            db.service_create(ctxt,
                              {'host': 'non_clustered_host',
                               'topic': constants.VOLUME_TOPIC,
                               'binary': constants.VOLUME_BINARY,
                               'created_at': timeutils.utcnow()}),
            # This service has no capabilities
            db.service_create(ctxt,
                              {'host': 'no_capabilities_host',
                               'topic': constants.VOLUME_TOPIC,
                               'binary': constants.VOLUME_BINARY,
                               'created_at': timeutils.utcnow()}),
        )

        capabilities = ((1, {'free_capacity_gb': 1000}),
                        # This is the capacity that will be selected for the
                        # cluster because is the one with the latest timestamp.
                        (3, {'free_capacity_gb': 2000}),
                        (2, {'free_capacity_gb': 3000}),
                        (1, {'free_capacity_gb': 4000}))

        for i in range(len(capabilities)):
            self.host_manager.update_service_capabilities(
                'volume', services[i].host, capabilities[i][1],
                services[i].cluster_name, capabilities[i][0])

        res = self.host_manager.get_all_backend_states(ctxt)
        result = {(s.cluster_name or s.host, s.free_capacity_gb) for s in res}
        expected = {(cluster_name + '#_pool0', 2000),
                    ('non_clustered_host#_pool0', 4000)}
        self.assertSetEqual(expected, result)

    @mock.patch('cinder.db.service_get_all')
    @mock.patch('cinder.objects.service.Service.is_up',
                new_callable=mock.PropertyMock)
    def test_get_all_backend_states(self, _mock_service_is_up,
                                    _mock_service_get_all):
        context = 'fake_context'
        timestamp = datetime.utcnow()
        topic = constants.VOLUME_TOPIC

        services = [
            dict(id=1, host='host1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 binary=None, deleted=False, created_at=None, modified_at=None,
                 report_count=0, deleted_at=None, disabled_reason=None,
                 uuid='a3a593da-7f8d-4bb7-8b4c-f2bc1e0b4824'),
            dict(id=2, host='host2', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 binary=None, deleted=False, created_at=None, modified_at=None,
                 report_count=0, deleted_at=None, disabled_reason=None,
                 uuid='4200b32b-0bf9-436c-86b2-0675f6ac218e'),
            dict(id=3, host='host3', topic='volume', disabled=False,
                 availability_zone='zone2', updated_at=timeutils.utcnow(),
                 binary=None, deleted=False, created_at=None, modified_at=None,
                 report_count=0, deleted_at=None, disabled_reason=None,
                 uuid='6d91e7f5-ca17-4e3b-bf4f-19ca77166dd7'),
            dict(id=4, host='host4', topic='volume', disabled=False,
                 availability_zone='zone3', updated_at=timeutils.utcnow(),
                 binary=None, deleted=False, created_at=None, modified_at=None,
                 report_count=0, deleted_at=None, disabled_reason=None,
                 uuid='18417850-2ca9-43d1-9619-ae16bfb0f655'),
        ]

        service_objs = []
        for db_service in services:
            service_obj = objects.Service()
            service_objs.append(objects.Service._from_db_object(context,
                                                                service_obj,
                                                                db_service))

        service_states = {
            'host1': dict(volume_backend_name='AAA',
                          total_capacity_gb=512, free_capacity_gb=200,
                          timestamp=timestamp, reserved_percentage=0,
                          provisioned_capacity_gb=312),
            'host2': dict(volume_backend_name='BBB',
                          total_capacity_gb=256, free_capacity_gb=100,
                          timestamp=timestamp, reserved_percentage=0,
                          provisioned_capacity_gb=156),
            'host3': dict(volume_backend_name='CCC',
                          total_capacity_gb=10000, free_capacity_gb=700,
                          timestamp=timestamp, reserved_percentage=0,
                          provisioned_capacity_gb=9300),
        }
        # First test: service.is_up is always True, host5 is disabled,
        # host4 has no capabilities
        self.host_manager.service_states = service_states
        _mock_service_get_all.return_value = services
        _mock_service_is_up.return_value = True
        _mock_warning = mock.Mock()
        host_manager.LOG.warning = _mock_warning

        # Get all states
        self.host_manager.get_all_backend_states(context)
        _mock_service_get_all.assert_called_with(context,
                                                 disabled=False,
                                                 frozen=False,
                                                 topic=topic)

        # verify that Service.is_up was called for each srv
        expected = [mock.call() for s in service_objs]
        self.assertEqual(expected, _mock_service_is_up.call_args_list)

        # Get backend_state_map and make sure we have the first 3 hosts
        backend_state_map = self.host_manager.backend_state_map
        self.assertEqual(3, len(backend_state_map))
        for i in range(3):
            volume_node = services[i]
            host = volume_node['host']
            test_service.TestService._compare(self, volume_node,
                                              backend_state_map[host].service)

        # Second test: Now service.is_up returns False for host3
        _mock_service_is_up.reset_mock()
        _mock_service_is_up.side_effect = [True, True, False, True]
        _mock_service_get_all.reset_mock()
        _mock_warning.reset_mock()

        # Get all states, make sure host 3 is reported as down
        self.host_manager.get_all_backend_states(context)
        _mock_service_get_all.assert_called_with(context,
                                                 disabled=False,
                                                 frozen=False,
                                                 topic=topic)

        self.assertEqual(expected, _mock_service_is_up.call_args_list)
        self.assertGreater(_mock_warning.call_count, 0)

        # Get backend_state_map and make sure we have the first 2 hosts (host3
        # is down, host4 is missing capabilities)
        backend_state_map = self.host_manager.backend_state_map
        self.assertEqual(2, len(backend_state_map))
        for i in range(2):
            volume_node = services[i]
            host = volume_node['host']
            test_service.TestService._compare(self, volume_node,
                                              backend_state_map[host].service)

    @mock.patch('cinder.db.service_get_all')
    @mock.patch('cinder.objects.service.Service.is_up',
                new_callable=mock.PropertyMock)
    def test_get_pools(self, _mock_service_is_up,
                       _mock_service_get_all):
        context = 'fake_context'
        timestamp = datetime.utcnow()

        services = [
            dict(id=1, host='host1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='a3a593da-7f8d-4bb7-8b4c-f2bc1e0b4824'),
            dict(id=2, host='host2@back1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='4200b32b-0bf9-436c-86b2-0675f6ac218e'),
            dict(id=3, host='host2@back2', topic='volume', disabled=False,
                 availability_zone='zone2', updated_at=timeutils.utcnow(),
                 uuid='6d91e7f5-ca17-4e3b-bf4f-19ca77166dd7'),
        ]

        mocked_service_states = {
            'host1': dict(volume_backend_name='AAA',
                          total_capacity_gb=512, free_capacity_gb=200,
                          timestamp=timestamp, reserved_percentage=0,
                          provisioned_capacity_gb=312),
            'host2@back1': dict(volume_backend_name='BBB',
                                total_capacity_gb=256, free_capacity_gb=100,
                                timestamp=timestamp, reserved_percentage=0,
                                provisioned_capacity_gb=156),
            'host2@back2': dict(volume_backend_name='CCC',
                                total_capacity_gb=10000, free_capacity_gb=700,
                                timestamp=timestamp, reserved_percentage=0,
                                provisioned_capacity_gb=9300),
        }

        _mock_service_get_all.return_value = services
        _mock_service_is_up.return_value = True
        _mock_warning = mock.Mock()
        host_manager.LOG.warn = _mock_warning

        with mock.patch.dict(self.host_manager.service_states,
                             mocked_service_states):
            res = self.host_manager.get_pools(context)

            # check if get_pools returns all 3 pools
            self.assertEqual(3, len(res))

            expected = [
                {
                    'name': 'host1#AAA',
                    'capabilities': {
                        'timestamp': timestamp,
                        'volume_backend_name': 'AAA',
                        'free_capacity_gb': 200,
                        'driver_version': None,
                        'total_capacity_gb': 512,
                        'reserved_percentage': 0,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'provisioned_capacity_gb': 312},
                },
                {
                    'name': 'host2@back1#BBB',
                    'capabilities': {
                        'timestamp': timestamp,
                        'volume_backend_name': 'BBB',
                        'free_capacity_gb': 100,
                        'driver_version': None,
                        'total_capacity_gb': 256,
                        'reserved_percentage': 0,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'provisioned_capacity_gb': 156},
                },
                {
                    'name': 'host2@back2#CCC',
                    'capabilities': {
                        'timestamp': timestamp,
                        'volume_backend_name': 'CCC',
                        'free_capacity_gb': 700,
                        'driver_version': None,
                        'total_capacity_gb': 10000,
                        'reserved_percentage': 0,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'provisioned_capacity_gb': 9300},
                }
            ]

            def sort_func(data):
                return data['name']

            self.assertEqual(len(expected), len(res))
            self.assertEqual(sorted(expected, key=sort_func),
                             sorted(res, key=sort_func))

    def test_get_usage(self):
        host = "host1@backend1"
        timestamp = 40000
        volume_stats1 = {'pools': [
                         {'pool_name': 'pool1',
                          'total_capacity_gb': 30.01,
                          'free_capacity_gb': 28.01,
                          'allocated_capacity_gb': 2.0,
                          'provisioned_capacity_gb': 2.0,
                          'max_over_subscription_ratio': '1.0',
                          'thin_provisioning_support': False,
                          'thick_provisioning_support': True,
                          'reserved_percentage': 5},
                         {'pool_name': 'pool2',
                          'total_capacity_gb': 20.01,
                          'free_capacity_gb': 18.01,
                          'allocated_capacity_gb': 2.0,
                          'provisioned_capacity_gb': 2.0,
                          'max_over_subscription_ratio': '2.0',
                          'thin_provisioning_support': True,
                          'thick_provisioning_support': False,
                          'reserved_percentage': 5}]}

        updated_pools1 = [{'pool_name': 'pool1',
                           'total_capacity_gb': 30.01,
                           'free_capacity_gb': 28.01,
                           'allocated_capacity_gb': 2.0,
                           'provisioned_capacity_gb': 2.0,
                           'max_over_subscription_ratio': '1.0',
                           'thin_provisioning_support': False,
                           'thick_provisioning_support': True,
                           'reserved_percentage': 5},
                          {'pool_name': 'pool2',
                           'total_capacity_gb': 20.01,
                           'free_capacity_gb': 18.01,
                           'allocated_capacity_gb': 2.0,
                           'provisioned_capacity_gb': 2.0,
                           'max_over_subscription_ratio': '2.0',
                           'thin_provisioning_support': True,
                           'thick_provisioning_support': False,
                           'reserved_percentage': 5}]

        volume_stats2 = {'pools': [
                         {'pool_name': 'pool1',
                          'total_capacity_gb': 30.01,
                          'free_capacity_gb': 28.01,
                          'allocated_capacity_gb': 2.0,
                          'provisioned_capacity_gb': 2.0,
                          'max_over_subscription_ratio': '2.0',
                          'thin_provisioning_support': True,
                          'thick_provisioning_support': False,
                          'reserved_percentage': 0},
                         {'pool_name': 'pool2',
                          'total_capacity_gb': 20.01,
                          'free_capacity_gb': 18.01,
                          'allocated_capacity_gb': 2.0,
                          'provisioned_capacity_gb': 2.0,
                          'max_over_subscription_ratio': '2.0',
                          'thin_provisioning_support': True,
                          'thick_provisioning_support': False,
                          'reserved_percentage': 5}]}

        updated_pools2 = [{'pool_name': 'pool1',
                           'total_capacity_gb': 30.01,
                           'free_capacity_gb': 28.01,
                           'allocated_capacity_gb': 2.0,
                           'provisioned_capacity_gb': 2.0,
                           'max_over_subscription_ratio': '2.0',
                           'thin_provisioning_support': True,
                           'thick_provisioning_support': False,
                           'reserved_percentage': 0}]

        expected1 = [
            {"name_to_id": 'host1@backend1#pool1',
             "type": "pool",
             "total": 30.01,
             "free": 28.01,
             "allocated": 2.0,
             "provisioned": 2.0,
             "virtual_free": 27.01,
             "reported_at": 40000},
            {"name_to_id": 'host1@backend1#pool2',
             "type": "pool",
             "total": 20.01,
             "free": 18.01,
             "allocated": 2.0,
             "provisioned": 2.0,
             "virtual_free": 37.02,
             "reported_at": 40000},
            {"name_to_id": 'host1@backend1',
             "type": "backend",
             "total": 50.02,
             "free": 46.02,
             "allocated": 4.0,
             "provisioned": 4.0,
             "virtual_free": 64.03,
             "reported_at": 40000}]

        expected2 = [
            {"name_to_id": 'host1@backend1#pool1',
             "type": "pool",
             "total": 30.01,
             "free": 28.01,
             "allocated": 2.0,
             "provisioned": 2.0,
             "virtual_free": 58.02,
             "reported_at": 40000},
            {"name_to_id": 'host1@backend1',
             "type": "backend",
             "total": 50.02,
             "free": 46.02,
             "allocated": 4.0,
             "provisioned": 4.0,
             "virtual_free": 95.04,
             "reported_at": 40000}]

        def sort_func(data):
            return data['name_to_id']

        res1 = self.host_manager._get_usage(volume_stats1,
                                            updated_pools1, host, timestamp)
        self.assertEqual(len(expected1), len(res1))
        self.assertEqual(sorted(expected1, key=sort_func),
                         sorted(res1, key=sort_func))

        res2 = self.host_manager._get_usage(volume_stats2,
                                            updated_pools2, host, timestamp)
        self.assertEqual(len(expected2), len(res2))
        self.assertEqual(sorted(expected2, key=sort_func),
                         sorted(res2, key=sort_func))

    @mock.patch('cinder.db.service_get_all')
    @mock.patch('cinder.objects.service.Service.is_up',
                new_callable=mock.PropertyMock)
    def test_get_pools_filter_name(self, _mock_service_is_up,
                                   _mock_service_get_all_by_topic):
        context = 'fake_context'

        services = [
            dict(id=1, host='host1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='a3a593da-7f8d-4bb7-8b4c-f2bc1e0b4824'),
            dict(id=2, host='host2@back1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='4200b32b-0bf9-436c-86b2-0675f6ac218e')
        ]

        mocked_service_states = {
            'host1': dict(volume_backend_name='AAA',
                          total_capacity_gb=512, free_capacity_gb=200,
                          timestamp=None, reserved_percentage=0,
                          provisioned_capacity_gb=312),
            'host2@back1': dict(volume_backend_name='BBB',
                                total_capacity_gb=256, free_capacity_gb=100,
                                timestamp=None, reserved_percentage=0,
                                provisioned_capacity_gb=156)
        }

        _mock_service_get_all_by_topic.return_value = services
        _mock_service_is_up.return_value = True
        _mock_warning = mock.Mock()
        host_manager.LOG.warn = _mock_warning

        with mock.patch.dict(self.host_manager.service_states,
                             mocked_service_states):
            filters = {'name': 'host1#AAA'}
            res = self.host_manager.get_pools(context, filters=filters)

            expected = [
                {
                    'name': 'host1#AAA',
                    'capabilities': {
                        'timestamp': None,
                        'volume_backend_name': 'AAA',
                        'free_capacity_gb': 200,
                        'driver_version': None,
                        'total_capacity_gb': 512,
                        'reserved_percentage': 0,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'provisioned_capacity_gb': 312},
                }
            ]

            self.assertEqual(expected, res)

    @mock.patch('cinder.scheduler.host_manager.HostManager.'
                '_choose_backend_filters')
    def test_get_pools_filtered_by_volume_type(self,
                                               _mock_choose_backend_filters):
        context = 'fake_context'
        filter_class = FakeFilterClass3
        _mock_choose_backend_filters.return_value = [filter_class]

        hosts = {
            'host1': {'volume_backend_name': 'AAA',
                      'total_capacity_gb': 512,
                      'free_capacity_gb': 200,
                      'timestamp': None,
                      'reserved_percentage': 0,
                      'provisioned_capacity_gb': 312},
            'host2@back1': {'volume_backend_name': 'BBB',
                            'total_capacity_gb': 256,
                            'free_capacity_gb': 100,
                            'timestamp': None,
                            'reserved_percentage': 0,
                            'provisioned_capacity_gb': 156}}
        mock_warning = mock.Mock()
        host_manager.LOG.warn = mock_warning
        mock_volume_type = {
            'volume_backend_name': 'AAA',
            'qos_specs': 'BBB',
        }
        host_manager1 = host_manager.HostManager()
        res = host_manager1._filter_pools_by_volume_type(context,
                                                         mock_volume_type,
                                                         hosts)
        expected = {'host1': {'volume_backend_name': 'AAA',
                              'total_capacity_gb': 512,
                              'free_capacity_gb': 200,
                              'timestamp': None, 'reserved_percentage': 0,
                              'provisioned_capacity_gb': 312}}

        self.assertEqual(expected, res)

    @mock.patch('cinder.db.service_get_all')
    @mock.patch('cinder.objects.service.Service.is_up',
                new_callable=mock.PropertyMock)
    def test_get_pools_filter_multiattach(self, _mock_service_is_up,
                                          _mock_service_get_all_by_topic):
        context = 'fake_context'

        services = [
            dict(id=1, host='host1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='a3a593da-7f8d-4bb7-8b4c-f2bc1e0b4824'),
            dict(id=2, host='host2@back1', topic='volume', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow(),
                 uuid='4200b32b-0bf9-436c-86b2-0675f6ac218e')
        ]

        mocked_service_states = {
            'host1': dict(volume_backend_name='AAA',
                          total_capacity_gb=512, free_capacity_gb=200,
                          timestamp=None, reserved_percentage=0,
                          multiattach=True),
            'host2@back1': dict(volume_backend_name='BBB',
                                total_capacity_gb=256, free_capacity_gb=100,
                                timestamp=None, reserved_percentage=0,
                                multiattach=False)
        }

        _mock_service_get_all_by_topic.return_value = services
        _mock_service_is_up.return_value = True
        _mock_warning = mock.Mock()
        host_manager.LOG.warn = _mock_warning

        with mock.patch.dict(self.host_manager.service_states,
                             mocked_service_states):
            filters_t = {'multiattach': 'true'}
            filters_f = {'multiattach': False}
            res_t = self.host_manager.get_pools(context, filters=filters_t)
            res_f = self.host_manager.get_pools(context, filters=filters_f)

            expected_t = [
                {
                    'name': 'host1#AAA',
                    'capabilities': {
                        'timestamp': None,
                        'volume_backend_name': 'AAA',
                        'free_capacity_gb': 200,
                        'driver_version': None,
                        'total_capacity_gb': 512,
                        'reserved_percentage': 0,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'multiattach': True},
                }
            ]
            expected_f = [
                {
                    'name': 'host2@back1#BBB',
                    'capabilities': {
                        'timestamp': None,
                        'volume_backend_name': 'BBB',
                        'free_capacity_gb': 100,
                        'driver_version': None,
                        'total_capacity_gb': 256,
                        'reserved_percentage': 0,
                        'vendor_name': None,
                        'storage_protocol': None,
                        'multiattach': False},
                }
            ]

            self.assertEqual(expected_t, res_t)
            self.assertEqual(expected_f, res_f)

    @ddt.data(
        (None, None, True),
        (None, 'value', False),
        ('cap', None, False),
        (False, 'True', False),
        (True, 'True', True),
        (True, True, True),
        (False, 'false', True),
        (1.1, '1.1', True),
        (0, '0', True),
        (1.1, '1.11', False),
        ('str', 'str', True),
        ('str1', 'str2', False),
        ('str', 'StR', False),
        ([], [], True),
        (['hdd', 'ssd'], ['ssd'], False),
        (['hdd', 'ssd'], ['ssd', 'hdd'], False),
        (['hdd', 'ssd'], "['hdd', 'ssd']", True),
        ({}, {}, True),
        ({'a': 'a', 'b': 'b'}, {'b': 'b', 'a': 'a'}, True),
        ({'a': 'a', 'b': 'b'}, {'b': 'b'}, False),
        ({'a': 'a'}, "{'a': 'a'}", True),
    )
    @ddt.unpack
    def test_equal_after_convert(self, cap, value, ret_value):
        self.assertEqual(ret_value,
                         self.host_manager._equal_after_convert(cap, value))


class BackendStateTestCase(test.TestCase):
    """Test case for BackendState class."""

    def test_update_from_volume_capability_nopool(self):
        fake_backend = host_manager.BackendState('be1', None)
        self.assertIsNone(fake_backend.free_capacity_gb)

        volume_capability = {'total_capacity_gb': 1024,
                             'free_capacity_gb': 512,
                             'provisioned_capacity_gb': 512,
                             'reserved_percentage': 0,
                             'timestamp': None}

        fake_backend.update_from_volume_capability(volume_capability)
        # Backend level stats remain uninitialized
        self.assertEqual(0, fake_backend.total_capacity_gb)
        self.assertIsNone(fake_backend.free_capacity_gb)
        # Pool stats has been updated
        self.assertEqual(1024, fake_backend.pools['_pool0'].total_capacity_gb)
        self.assertEqual(512, fake_backend.pools['_pool0'].free_capacity_gb)
        self.assertEqual(512,
                         fake_backend.pools['_pool0'].provisioned_capacity_gb)

        # Test update for existing host state
        volume_capability.update(dict(total_capacity_gb=1000))
        fake_backend.update_from_volume_capability(volume_capability)
        self.assertEqual(1000, fake_backend.pools['_pool0'].total_capacity_gb)

        # Test update for existing host state with different backend name
        volume_capability.update(dict(volume_backend_name='magic'))
        fake_backend.update_from_volume_capability(volume_capability)
        self.assertEqual(1000, fake_backend.pools['magic'].total_capacity_gb)
        self.assertEqual(512, fake_backend.pools['magic'].free_capacity_gb)
        self.assertEqual(512,
                         fake_backend.pools['magic'].provisioned_capacity_gb)
        # 'pool0' becomes nonactive pool, and is deleted
        self.assertRaises(KeyError, lambda: fake_backend.pools['pool0'])

    def test_update_from_volume_capability_with_pools(self):
        fake_backend = host_manager.BackendState('host1', None)
        self.assertIsNone(fake_backend.free_capacity_gb)
        capability = {
            'volume_backend_name': 'Local iSCSI',
            'vendor_name': 'OpenStack',
            'driver_version': '1.0.1',
            'storage_protocol': 'iSCSI',
            'pools': [
                {'pool_name': '1st pool',
                 'total_capacity_gb': 500,
                 'free_capacity_gb': 230,
                 'allocated_capacity_gb': 270,
                 'provisioned_capacity_gb': 270,
                 'QoS_support': 'False',
                 'reserved_percentage': 0,
                 'dying_disks': 100,
                 'super_hero_1': 'spider-man',
                 'super_hero_2': 'flash',
                 'super_hero_3': 'neoncat',
                 },
                {'pool_name': '2nd pool',
                 'total_capacity_gb': 1024,
                 'free_capacity_gb': 1024,
                 'allocated_capacity_gb': 0,
                 'provisioned_capacity_gb': 0,
                 'QoS_support': 'False',
                 'reserved_percentage': 0,
                 'dying_disks': 200,
                 'super_hero_1': 'superman',
                 'super_hero_2': 'Hulk',
                 }
            ],
            'timestamp': None,
        }

        fake_backend.update_from_volume_capability(capability)

        self.assertEqual('Local iSCSI', fake_backend.volume_backend_name)
        self.assertEqual('iSCSI', fake_backend.storage_protocol)
        self.assertEqual('OpenStack', fake_backend.vendor_name)
        self.assertEqual('1.0.1', fake_backend.driver_version)

        # Backend level stats remain uninitialized
        self.assertEqual(0, fake_backend.total_capacity_gb)
        self.assertIsNone(fake_backend.free_capacity_gb)
        # Pool stats has been updated
        self.assertEqual(2, len(fake_backend.pools))

        self.assertEqual(500, fake_backend.pools['1st pool'].total_capacity_gb)
        self.assertEqual(230, fake_backend.pools['1st pool'].free_capacity_gb)
        self.assertEqual(
            270, fake_backend.pools['1st pool'].provisioned_capacity_gb)
        self.assertEqual(
            1024, fake_backend.pools['2nd pool'].total_capacity_gb)
        self.assertEqual(1024, fake_backend.pools['2nd pool'].free_capacity_gb)
        self.assertEqual(
            0, fake_backend.pools['2nd pool'].provisioned_capacity_gb)

        capability = {
            'volume_backend_name': 'Local iSCSI',
            'vendor_name': 'OpenStack',
            'driver_version': '1.0.2',
            'storage_protocol': 'iSCSI',
            'pools': [
                {'pool_name': '3rd pool',
                 'total_capacity_gb': 10000,
                 'free_capacity_gb': 10000,
                 'allocated_capacity_gb': 0,
                 'provisioned_capacity_gb': 0,
                 'QoS_support': 'False',
                 'reserved_percentage': 0,
                 },
            ],
            'timestamp': None,
        }

        # test update BackendState Record
        fake_backend.update_from_volume_capability(capability)

        self.assertEqual('1.0.2', fake_backend.driver_version)

        # Non-active pool stats has been removed
        self.assertEqual(1, len(fake_backend.pools))

        self.assertRaises(KeyError, lambda: fake_backend.pools['1st pool'])
        self.assertRaises(KeyError, lambda: fake_backend.pools['2nd pool'])

        self.assertEqual(10000,
                         fake_backend.pools['3rd pool'].total_capacity_gb)
        self.assertEqual(10000,
                         fake_backend.pools['3rd pool'].free_capacity_gb)
        self.assertEqual(
            0, fake_backend.pools['3rd pool'].provisioned_capacity_gb)

    def test_update_from_volume_infinite_capability(self):
        fake_backend = host_manager.BackendState('host1', None)
        self.assertIsNone(fake_backend.free_capacity_gb)

        volume_capability = {'total_capacity_gb': 'infinite',
                             'free_capacity_gb': 'infinite',
                             'reserved_percentage': 0,
                             'timestamp': None}

        fake_backend.update_from_volume_capability(volume_capability)
        # Backend level stats remain uninitialized
        self.assertEqual(0, fake_backend.total_capacity_gb)
        self.assertIsNone(fake_backend.free_capacity_gb)
        # Pool stats has been updated
        self.assertEqual(
            'infinite',
            fake_backend.pools['_pool0'].total_capacity_gb)
        self.assertEqual(
            'infinite',
            fake_backend.pools['_pool0'].free_capacity_gb)

    def test_update_from_volume_unknown_capability(self):
        fake_backend = host_manager.BackendState('host1', None)
        self.assertIsNone(fake_backend.free_capacity_gb)

        volume_capability = {'total_capacity_gb': 'infinite',
                             'free_capacity_gb': 'unknown',
                             'reserved_percentage': 0,
                             'timestamp': None}

        fake_backend.update_from_volume_capability(volume_capability)
        # Backend level stats remain uninitialized
        self.assertEqual(0, fake_backend.total_capacity_gb)
        self.assertIsNone(fake_backend.free_capacity_gb)
        # Pool stats has been updated
        self.assertEqual(
            'infinite',
            fake_backend.pools['_pool0'].total_capacity_gb)
        self.assertEqual(
            'unknown',
            fake_backend.pools['_pool0'].free_capacity_gb)

    def test_update_from_empty_volume_capability(self):
        fake_backend = host_manager.BackendState('host1', None)

        vol_cap = {'timestamp': None}

        fake_backend.update_from_volume_capability(vol_cap)
        self.assertEqual(0, fake_backend.total_capacity_gb)
        self.assertIsNone(fake_backend.free_capacity_gb)
        # Pool stats has been updated
        self.assertEqual(0,
                         fake_backend.pools['_pool0'].total_capacity_gb)
        self.assertEqual(0,
                         fake_backend.pools['_pool0'].free_capacity_gb)
        self.assertEqual(0,
                         fake_backend.pools['_pool0'].provisioned_capacity_gb)


class PoolStateTestCase(test.TestCase):
    """Test case for BackendState class."""

    def test_update_from_volume_capability(self):
        fake_pool = host_manager.PoolState('host1', None, None, 'pool0')
        self.assertIsNone(fake_pool.free_capacity_gb)

        volume_capability = {'total_capacity_gb': 1024,
                             'free_capacity_gb': 512,
                             'reserved_percentage': 0,
                             'provisioned_capacity_gb': 512,
                             'timestamp': None,
                             'cap1': 'val1',
                             'cap2': 'val2'}

        fake_pool.update_from_volume_capability(volume_capability)
        self.assertEqual('host1#pool0', fake_pool.host)
        self.assertEqual('pool0', fake_pool.pool_name)
        self.assertEqual(1024, fake_pool.total_capacity_gb)
        self.assertEqual(512, fake_pool.free_capacity_gb)
        self.assertEqual(512,
                         fake_pool.provisioned_capacity_gb)

        self.assertDictEqual(volume_capability, dict(fake_pool.capabilities))



# Copyright 2017 Inspur Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import random

from eventlet import greenthread
from oslo_concurrency import processutils
from oslo_log import log as logging
from oslo_utils import excutils
import six

from cinder import exception
from cinder.i18n import _
from cinder.objects import fields
from cinder import ssh_utils
from cinder import utils as cinder_utils
from cinder.volume.drivers.inspur.instorage import instorage_const

LOG = logging.getLogger(__name__)


class InStorageMCSReplicationManager(object):

    def __init__(self, driver, replication_target=None, target_assistant=None):
        self.sshpool = None
        self.driver = driver
        self.target = replication_target
        self.target_assistant = target_assistant(self._run_ssh)
        self._local_assistant = self.driver._local_backend_assistant
        self.async_m = InStorageMCSReplicationAsyncCopy(
            self.driver, replication_target, self.target_assistant)
        self.sync_m = InStorageMCSReplicationSyncCopy(
            self.driver, replication_target, self.target_assistant)

    def _run_ssh(self, cmd_list, check_exit_code=True, attempts=1):
        cinder_utils.check_ssh_injection(cmd_list)
        command = ' '. join(cmd_list)

        if not self.sshpool:
            self.sshpool = ssh_utils.SSHPool(
                self.target.get('san_ip'),
                self.target.get('san_ssh_port', 22),
                self.target.get('ssh_conn_timeout', 30),
                self.target.get('san_login'),
                password=self.target.get('san_password'),
                privatekey=self.target.get('san_private_key', ''),
                min_size=self.target.get('ssh_min_pool_conn', 1),
                max_size=self.target.get('ssh_max_pool_conn', 5),)
        last_exception = None
        try:
            with self.sshpool.item() as ssh:
                while attempts > 0:
                    attempts -= 1
                    try:
                        return processutils.ssh_execute(
                            ssh, command, check_exit_code=check_exit_code)
                    except Exception as e:
                        LOG.error(e)
                        last_exception = e
                        greenthread.sleep(random.randint(20, 500) / 100.0)
                try:
                    raise processutils.ProcessExecutionError(
                        exit_code=last_exception.exit_code,
                        stdout=last_exception.stdout,
                        stderr=last_exception.stderr,
                        cmd=last_exception.cmd)
                except AttributeError:
                    raise processutils.ProcessExecutionError(
                        exit_code=-1, stdout="",
                        stderr="Error running SSH command",
                        cmd=command)
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("Error running SSH command: %s", command)

    def get_target_assistant(self):
        return self.target_assistant

    def get_replica_obj(self, rep_type):
        if rep_type == instorage_const.ASYNC:
            return self.async_m
        elif rep_type == instorage_const.SYNC:
            return self.sync_m
        else:
            return None

    def _partnership_validate_create(self, client, remote_name, remote_ip):
        try:
            partnership_info = client.get_partnership_info(remote_name)
            if not partnership_info:
                candidate_info = client.get_partnershipcandidate_info(
                    remote_name)
                if candidate_info:
                    client.mkfcpartnership(remote_name)
                else:
                    client.mkippartnership(remote_ip)
                partnership_info = client.get_partnership_info(remote_name)
            if partnership_info['partnership'] != 'fully_configured':
                client.chpartnership(partnership_info['id'])
        except Exception:
            msg = (_('Unable to establish the partnership with '
                     'the InStorage cluster %s.') % remote_name)
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

    def establish_target_partnership(self):
        local_system_info = self._local_assistant.get_system_info()
        target_system_info = self.target_assistant.get_system_info()
        local_system_name = local_system_info['system_name']
        target_system_name = target_system_info['system_name']
        local_ip = self.driver.configuration.safe_get('san_ip')
        target_ip = self.target.get('san_ip')
        # Establish partnership only when the local system and the replication
        # target system is different.
        if target_system_name != local_system_name:
            self._partnership_validate_create(self._local_assistant,
                                              target_system_name, target_ip)
            self._partnership_validate_create(self.target_assistant,
                                              local_system_name, local_ip)


class InStorageMCSReplication(object):

    def __init__(self, asynccopy, driver,
                 replication_target=None, target_assistant=None):

        self.asynccopy = asynccopy
        self.driver = driver
        self.target = replication_target or {}
        self.target_assistant = target_assistant

    @cinder_utils.trace
    def volume_replication_setup(self, context, vref):
        target_vol_name = instorage_const.REPLICA_AUX_VOL_PREFIX + vref.name
        try:
            attr = self.target_assistant.get_vdisk_attributes(target_vol_name)
            if not attr:
                opts = self.driver._get_vdisk_params(vref.volume_type_id)
                pool = self.target.get('pool_name')
                src_attr = self.driver._assistant.get_vdisk_attributes(
                    vref.name)
                opts['iogrp'] = src_attr['IO_group_id']
                self.target_assistant.create_vdisk(target_vol_name,
                                                   six.text_type(vref['size']),
                                                   'gb', pool, opts)

            system_info = self.target_assistant.get_system_info()
            self.driver._assistant.create_relationship(
                vref.name, target_vol_name, system_info.get('system_name'),
                self.asynccopy)
        except Exception as e:
            msg = (_("Unable to set up copy mode replication for %(vol)s. "
                     "Exception: %(err)s.") % {'vol': vref.id, 'err': e})
            LOG.error(msg)
            raise exception.VolumeDriverException(message=msg)

    @cinder_utils.trace
    def failover_volume_host(self, context, vref):
        target_vol = instorage_const.REPLICA_AUX_VOL_PREFIX + vref.name

        try:
            rel_info = self.target_assistant.get_relationship_info(target_vol)
            # Reverse the role of the primary and secondary volumes
            self.target_assistant.switch_relationship(rel_info['name'])
            return {'replication_status': fields.ReplicationStatus.FAILED_OVER}
        except Exception as e:
            LOG.exception('Unable to fail-over the volume %(id)s to the '
                          'secondary back-end by switchrcrelationship '
                          'command.', {"id": vref.id})
            # If the switch command fail, try to make the aux volume
            # writeable again.
            try:
                self.target_assistant.stop_relationship(target_vol,
                                                        access=True)
                return {
                    'replication_status': fields.ReplicationStatus.FAILED_OVER}
            except Exception as e:
                msg = (_('Unable to fail-over the volume %(id)s to the '
                         'secondary back-end, error: %(error)s') %
                       {"id": vref.id, "error": six.text_type(e)})
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)

    def replication_failback(self, volume):
        tgt_volume = instorage_const.REPLICA_AUX_VOL_PREFIX + volume.name
        rel_info = self.target_assistant.get_relationship_info(tgt_volume)
        if rel_info:
            try:
                self.target_assistant.switch_relationship(rel_info['name'],
                                                          aux=False)
                return {'replication_status': fields.ReplicationStatus.ENABLED,
                        'status': 'available'}
            except Exception as e:
                msg = (_('Unable to fail-back the volume:%(vol)s to the '
                         'master back-end, error:%(error)s') %
                       {"vol": volume.name, "error": six.text_type(e)})
                LOG.error(msg)
                raise exception.VolumeDriverException(message=msg)


class InStorageMCSReplicationAsyncCopy(InStorageMCSReplication):
    """Support for InStorage/MCS async copy mode replication.

    Async Copy establishes a Async Copy relationship between
    two volumes of equal size. The volumes in a Async Copy relationship
    are referred to as the master (source) volume and the auxiliary
    (target) volume. This mode is dedicated to the asynchronous volume
    replication.
    """

    def __init__(self, driver, replication_target=None, target_assistant=None):
        super(InStorageMCSReplicationAsyncCopy, self).__init__(
            True, driver, replication_target, target_assistant)


class InStorageMCSReplicationSyncCopy(InStorageMCSReplication):
    """Support for InStorage/MCS sync copy mode replication.

    Sync Copy establishes a Sync Copy relationship between
    two volumes of equal size. The volumes in a Sync Copy relationship
    are referred to as the master (source) volume and the auxiliary
    (target) volume.
    """

    def __init__(self, driver, replication_target=None, target_assistant=None):
        super(InStorageMCSReplicationSyncCopy, self).__init__(
            False, driver, replication_target, target_assistant)



# Copyright 2014 - 2017, 2019  StorPool
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import re
import sys

import ddt
import mock
from oslo_utils import units
import six


fakeStorPool = mock.Mock()
fakeStorPool.spopenstack = mock.Mock()
fakeStorPool.spapi = mock.Mock()
fakeStorPool.spconfig = mock.Mock()
fakeStorPool.sptypes = mock.Mock()
sys.modules['storpool'] = fakeStorPool


from cinder import exception
from cinder import test
from cinder.volume import configuration as conf
from cinder.volume.drivers import storpool as driver


volume_types = {
    1: {},
    2: {'storpool_template': 'ssd'},
    3: {'storpool_template': 'hdd'}
}
volumes = {}
snapshots = {}


def MockExtraSpecs(vtype):
    return volume_types[vtype]


def mock_volume_types(f):
    def _types_inner_inner1(inst, *args, **kwargs):
        @mock.patch('cinder.volume.volume_types.get_volume_type_extra_specs',
                    new=MockExtraSpecs)
        def _types_inner_inner2():
            return f(inst, *args, **kwargs)

        return _types_inner_inner2()

    return _types_inner_inner1


def volumeName(vid):
    return 'os--volume--{id}'.format(id=vid)


def snapshotName(vtype, vid):
    return 'os--snap--{t}--{id}'.format(t=vtype, id=vid)


class MockDisk(object):
    def __init__(self, diskId):
        self.id = diskId
        self.generationLeft = -1
        self.agCount = 14
        self.agFree = 12
        self.agAllocated = 1


class MockVolume(object):
    def __init__(self, v):
        self.name = v['name']


class MockTemplate(object):
    def __init__(self, name):
        self.name = name


class MockApiError(Exception):
    def __init__(self, msg):
        super(MockApiError, self).__init__(msg)


class MockAPI(object):
    def __init__(self):
        self._disks = {diskId: MockDisk(diskId) for diskId in (1, 2, 3, 4)}
        self._disks[3].generationLeft = 42

        self._templates = [MockTemplate(name) for name in ('ssd', 'hdd')]

    def setlog(self, log):
        self._log = log

    def disksList(self):
        return self._disks

    def snapshotCreate(self, vname, snap):
        snapshots[snap['name']] = dict(volumes[vname])

    def snapshotDelete(self, name):
        del snapshots[name]

    def volumeCreate(self, v):
        if v['name'] in volumes:
            raise MockApiError('volume already exists')
        volumes[v['name']] = v

    def volumeDelete(self, name):
        del volumes[name]

    def volumesList(self):
        return [MockVolume(v[1]) for v in volumes.items()]

    def volumeTemplatesList(self):
        return self._templates

    def volumesReassign(self, json):
        pass

    def volumeUpdate(self, name, data):
        if 'size' in data:
            volumes[name]['size'] = data['size']

        if 'rename' in data and data['rename'] != name:
            volumes[data['rename']] = volumes[name]
            del volumes[name]


class MockAttachDB(object):
    def __init__(self, log):
        self._api = MockAPI()

    def api(self):
        return self._api

    def volumeName(self, vid):
        return volumeName(vid)

    def snapshotName(self, vtype, vid):
        return snapshotName(vtype, vid)


def MockVolumeUpdateDesc(size):
    return {'size': size}


def MockSPConfig(section = 's01'):
    res = {}
    m = re.match('^s0*([A-Za-z0-9]+)$', section)
    if m:
        res['SP_OURID'] = m.group(1)
    return res


fakeStorPool.spapi.ApiError = MockApiError
fakeStorPool.spconfig.SPConfig = MockSPConfig
fakeStorPool.spopenstack.AttachDB = MockAttachDB
fakeStorPool.sptypes.VolumeUpdateDesc = MockVolumeUpdateDesc


@ddt.ddt
class StorPoolTestCase(test.TestCase):

    def setUp(self):
        super(StorPoolTestCase, self).setUp()

        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.volume_backend_name = 'storpool_test'
        self.cfg.storpool_template = None
        self.cfg.storpool_replication = 3

        mock_exec = mock.Mock()
        mock_exec.return_value = ('', '')

        self.driver = driver.StorPoolDriver(execute=mock_exec,
                                            configuration=self.cfg)
        self.driver.check_for_setup_error()

    @ddt.data(
        (5, TypeError),
        ({'no-host': None}, KeyError),
        ({'host': 'sbad'}, driver.StorPoolConfigurationInvalid),
        ({'host': 's01'}, None),
        ({'host': 'none'}, None),
    )
    @ddt.unpack
    def test_validate_connector(self, conn, exc):
        if exc is None:
            self.assertTrue(self.driver.validate_connector(conn))
        else:
            self.assertRaises(exc,
                              self.driver.validate_connector,
                              conn)

    @ddt.data(
        (5, TypeError),
        ({'no-host': None}, KeyError),
        ({'host': 'sbad'}, driver.StorPoolConfigurationInvalid),
    )
    @ddt.unpack
    def test_initialize_connection_bad(self, conn, exc):
        self.assertRaises(exc,
                          self.driver.initialize_connection,
                          None, conn)

    @ddt.data(
        (1, '42', 's01'),
        (2, '616', 's02'),
        (65, '1610', 'none'),
    )
    @ddt.unpack
    def test_initialize_connection_good(self, cid, hid, name):
        c = self.driver.initialize_connection({'id': hid}, {'host': name})
        self.assertEqual('storpool', c['driver_volume_type'])
        self.assertDictEqual({'client_id': cid, 'volume': hid}, c['data'])

    def test_noop_functions(self):
        self.driver.terminate_connection(None, None)
        self.driver.create_export(None, None, {})
        self.driver.remove_export(None, None)

    def test_stats(self):
        stats = self.driver.get_volume_stats(refresh=True)
        self.assertEqual('StorPool', stats['vendor_name'])
        self.assertEqual('storpool', stats['storage_protocol'])
        self.assertListEqual(['default', 'template_hdd', 'template_ssd'],
                             sorted([p['pool_name'] for p in stats['pools']]))
        r = re.compile(r'^template_([A-Za-z0-9_]+)$')
        for pool in stats['pools']:
            self.assertEqual(21, pool['total_capacity_gb'])
            self.assertEqual(5, int(pool['free_capacity_gb']))

            self.assertTrue(pool['multiattach'])
            self.assertFalse(pool['QoS_support'])
            self.assertFalse(pool['thick_provisioning_support'])
            self.assertTrue(pool['thin_provisioning_support'])

            if pool['pool_name'] != 'default':
                m = r.match(pool['pool_name'])
                self.assertIsNotNone(m)
                self.assertIsNotNone(m.group(1))
                self.assertEqual(m.group(1), pool['storpool_template'])

    def assertVolumeNames(self, names):
        self.assertListEqual(sorted([volumeName(n) for n in names]),
                             sorted(volumes.keys()))

    @mock_volume_types
    def test_create_delete_volume(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        self.driver.create_volume({'id': '1', 'name': 'v1', 'size': 1,
                                   'volume_type': None})
        six.assertCountEqual(self, [volumeName('1')], volumes.keys())
        self.assertVolumeNames(('1',))
        v = volumes[volumeName('1')]
        self.assertEqual(1 * units.Gi, v['size'])
        self.assertNotIn('template', v.keys())
        self.assertEqual(3, v['replication'])

        caught = False
        try:
            self.driver.create_volume({'id': '1', 'name': 'v1', 'size': 0,
                                       'volume_type': None})
        except exception.VolumeBackendAPIException:
            caught = True
        self.assertTrue(caught)

        self.driver.delete_volume({'id': '1'})
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)

        self.driver.create_volume({'id': '1', 'name': 'v1', 'size': 2,
                                   'volume_type': None})
        self.assertVolumeNames(('1',))
        v = volumes[volumeName('1')]
        self.assertEqual(2 * units.Gi, v['size'])
        self.assertNotIn('template', v.keys())
        self.assertEqual(3, v['replication'])

        self.driver.create_volume({'id': '2', 'name': 'v2', 'size': 3,
                                   'volume_type': {'id': 1}})
        self.assertVolumeNames(('1', '2'))
        v = volumes[volumeName('2')]
        self.assertEqual(3 * units.Gi, v['size'])
        self.assertNotIn('template', v.keys())
        self.assertEqual(3, v['replication'])

        self.driver.create_volume({'id': '3', 'name': 'v2', 'size': 4,
                                   'volume_type': {'id': 2}})
        self.assertVolumeNames(('1', '2', '3'))
        v = volumes[volumeName('3')]
        self.assertEqual(4 * units.Gi, v['size'])
        self.assertEqual('ssd', v['template'])
        self.assertNotIn('replication', v.keys())

        self.driver.create_volume({'id': '4', 'name': 'v2', 'size': 5,
                                   'volume_type': {'id': 3}})
        self.assertVolumeNames(('1', '2', '3', '4'))
        v = volumes[volumeName('4')]
        self.assertEqual(5 * units.Gi, v['size'])
        self.assertEqual('hdd', v['template'])
        self.assertNotIn('replication', v.keys())

        # Make sure the dictionary is not corrupted somehow...
        v = volumes[volumeName('1')]
        self.assertEqual(2 * units.Gi, v['size'])
        self.assertNotIn('template', v.keys())
        self.assertEqual(3, v['replication'])

        for vid in ('1', '2', '3', '4'):
            self.driver.delete_volume({'id': vid})
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @mock_volume_types
    def test_update_migrated_volume(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        # Create two volumes
        self.driver.create_volume({'id': '1', 'name': 'v1', 'size': 1,
                                   'volume_type': None})
        self.driver.create_volume({'id': '2', 'name': 'v2', 'size': 1,
                                   'volume_type': None})
        six.assertCountEqual(self,
                             [volumeName('1'), volumeName('2')],
                             volumes.keys())
        self.assertVolumeNames(('1', '2',))

        # Failure: the "migrated" volume does not even exist
        res = self.driver.update_migrated_volume(None, {'id': '1'},
                                                 {'id': '3', '_name_id': '1'},
                                                 'available')
        self.assertDictEqual({'_name_id': '1'}, res)

        # Failure: a volume with the original volume's name already exists
        res = self.driver.update_migrated_volume(None, {'id': '1'},
                                                 {'id': '2', '_name_id': '1'},
                                                 'available')
        self.assertDictEqual({'_name_id': '1'}, res)

        # Success: rename the migrated volume to match the original
        res = self.driver.update_migrated_volume(None, {'id': '3'},
                                                 {'id': '2', '_name_id': '3'},
                                                 'available')
        self.assertDictEqual({'_name_id': None}, res)
        six.assertCountEqual(self,
                             [volumeName('1'), volumeName('3')],
                             volumes.keys())
        self.assertVolumeNames(('1', '3',))

        for vid in ('1', '3'):
            self.driver.delete_volume({'id': vid})
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    def test_clone_extend_volume(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        self.driver.create_volume({'id': '1', 'name': 'v1', 'size': 1,
                                   'volume_type': None})
        self.assertVolumeNames(('1',))
        self.driver.extend_volume({'id': '1'}, 2)
        self.assertEqual(2 * units.Gi, volumes[volumeName('1')]['size'])

        self.driver.create_cloned_volume({'id': '2', 'name': 'clo', 'size': 3},
                                         {'id': 1})
        self.assertVolumeNames(('1', '2'))
        self.assertDictEqual({}, snapshots)
        # Note: this would not be true in a real environment (the snapshot will
        # have been deleted, the volume would have no parent), but with this
        # fake implementation it helps us make sure that the second volume was
        # created with the proper options.
        self.assertEqual(volumes[volumeName('2')]['parent'],
                         snapshotName('clone', '2'))

        self.driver.delete_volume({'id': 1})
        self.driver.delete_volume({'id': 2})

        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @mock_volume_types
    def test_config_replication(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        save_repl = self.driver.configuration.storpool_replication

        self.driver.configuration.storpool_replication = 3
        stats = self.driver.get_volume_stats(refresh=True)
        pool = stats['pools'][0]
        self.assertEqual(21, pool['total_capacity_gb'])
        self.assertEqual(5, int(pool['free_capacity_gb']))

        self.driver.create_volume({'id': 'cfgrepl1', 'name': 'v1', 'size': 1,
                                   'volume_type': None})
        self.assertVolumeNames(('cfgrepl1',))
        v = volumes[volumeName('cfgrepl1')]
        self.assertEqual(3, v['replication'])
        self.assertNotIn('template', v)
        self.driver.delete_volume({'id': 'cfgrepl1'})

        self.driver.configuration.storpool_replication = 2
        stats = self.driver.get_volume_stats(refresh=True)
        pool = stats['pools'][0]
        self.assertEqual(21, pool['total_capacity_gb'])
        self.assertEqual(8, int(pool['free_capacity_gb']))

        self.driver.create_volume({'id': 'cfgrepl2', 'name': 'v1', 'size': 1,
                                   'volume_type': None})
        self.assertVolumeNames(('cfgrepl2',))
        v = volumes[volumeName('cfgrepl2')]
        self.assertEqual(2, v['replication'])
        self.assertNotIn('template', v)
        self.driver.delete_volume({'id': 'cfgrepl2'})

        self.driver.create_volume({'id': 'cfgrepl3', 'name': 'v1', 'size': 1,
                                   'volume_type': {'id': 2}})
        self.assertVolumeNames(('cfgrepl3',))
        v = volumes[volumeName('cfgrepl3')]
        self.assertNotIn('replication', v)
        self.assertEqual('ssd', v['template'])
        self.driver.delete_volume({'id': 'cfgrepl3'})

        self.driver.configuration.storpool_replication = save_repl

        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @mock_volume_types
    def test_config_template(self):
        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

        save_template = self.driver.configuration.storpool_template

        self.driver.configuration.storpool_template = None

        self.driver.create_volume({'id': 'cfgtempl1', 'name': 'v1', 'size': 1,
                                   'volume_type': None})
        self.assertVolumeNames(('cfgtempl1',))
        v = volumes[volumeName('cfgtempl1')]
        self.assertEqual(3, v['replication'])
        self.assertNotIn('template', v)
        self.driver.delete_volume({'id': 'cfgtempl1'})

        self.driver.create_volume({'id': 'cfgtempl2', 'name': 'v1', 'size': 1,
                                   'volume_type': {'id': 2}})
        self.assertVolumeNames(('cfgtempl2',))
        v = volumes[volumeName('cfgtempl2')]
        self.assertNotIn('replication', v)
        self.assertEqual('ssd', v['template'])
        self.driver.delete_volume({'id': 'cfgtempl2'})

        self.driver.configuration.storpool_template = 'hdd'

        self.driver.create_volume({'id': 'cfgtempl3', 'name': 'v1', 'size': 1,
                                   'volume_type': None})
        self.assertVolumeNames(('cfgtempl3',))
        v = volumes[volumeName('cfgtempl3')]
        self.assertNotIn('replication', v)
        self.assertEqual('hdd', v['template'])
        self.driver.delete_volume({'id': 'cfgtempl3'})

        self.driver.create_volume({'id': 'cfgtempl4', 'name': 'v1', 'size': 1,
                                   'volume_type': {'id': 2}})
        self.assertVolumeNames(('cfgtempl4',))
        v = volumes[volumeName('cfgtempl4')]
        self.assertNotIn('replication', v)
        self.assertEqual('ssd', v['template'])
        self.driver.delete_volume({'id': 'cfgtempl4'})

        self.driver.configuration.storpool_template = save_template

        self.assertVolumeNames([])
        self.assertDictEqual({}, volumes)
        self.assertDictEqual({}, snapshots)

    @ddt.data(
        # No volume type at all: 'default'
        ('default', None),
        # No storpool_template in the type extra specs: 'default'
        ('default', {'id': 1}),
        # An actual template specified: 'template_*'
        ('template_ssd', {'id': 2}),
        ('template_hdd', {'id': 3}),
    )
    @ddt.unpack
    @mock_volume_types
    def test_get_pool(self, pool, volume_type):
        self.assertEqual(pool,
                         self.driver.get_pool({
                             'volume_type': volume_type
                         }))



#!/usr/bin/env python
# Copyright (c) 2011 X.commerce, a business unit of eBay Inc.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Interactive shell based on Django:
#
# Copyright (c) 2005, the Lawrence Journal-World
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#     3. Neither the name of Django nor the names of its contributors may be
#        used to endorse or promote products derived from this software without
#        specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


"""CLI interface for cinder management."""

from __future__ import print_function

try:
    import collections.abc as collections
except ImportError:
    import collections

import logging as python_logging
import prettytable
import sys
import time

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import migration
from oslo_log import log as logging
from oslo_utils import timeutils

# Need to register global_opts
from cinder.backup import rpcapi as backup_rpcapi
from cinder.common import config  # noqa
from cinder.common import constants
from cinder import context
from cinder import db
from cinder.db import migration as db_migration
from cinder.db.sqlalchemy import api as db_api
from cinder.db.sqlalchemy import models
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base as ovo_base
from cinder import rpc
from cinder.scheduler import rpcapi as scheduler_rpcapi
from cinder import version
from cinder.volume import rpcapi as volume_rpcapi
from cinder.volume import volume_utils


CONF = cfg.CONF

LOG = logging.getLogger(__name__)

RPC_VERSIONS = {
    'cinder-scheduler': scheduler_rpcapi.SchedulerAPI.RPC_API_VERSION,
    'cinder-volume': volume_rpcapi.VolumeAPI.RPC_API_VERSION,
    'cinder-backup': backup_rpcapi.BackupAPI.RPC_API_VERSION,
}

OVO_VERSION = ovo_base.OBJ_VERSIONS.get_current()


def _get_non_shared_target_hosts(ctxt):
    hosts = []
    numvols_needing_update = 0
    rpc.init(CONF)
    rpcapi = volume_rpcapi.VolumeAPI()

    services = objects.ServiceList.get_all_by_topic(ctxt,
                                                    constants.VOLUME_TOPIC)
    for service in services:
        capabilities = rpcapi.get_capabilities(ctxt, service.host, True)
        # Select only non iSCSI connections and iSCSI that are explicit
        if (capabilities.get('storage_protocol') != 'iSCSI' or
                not capabilities.get('shared_targets', True)):
            hosts.append(service.host)
            numvols_needing_update += db_api.model_query(
                ctxt, models.Volume).filter_by(
                    shared_targets=True,
                    service_uuid=service.uuid).count()
    return hosts, numvols_needing_update


# Decorators for actions
def args(*args, **kwargs):
    def _decorator(func):
        func.__dict__.setdefault('args', []).insert(0, (args, kwargs))
        return func
    return _decorator


class ShellCommands(object):
    def bpython(self):
        """Runs a bpython shell.

        Falls back to Ipython/python shell if unavailable
        """
        self.run('bpython')

    def ipython(self):
        """Runs an Ipython shell.

        Falls back to Python shell if unavailable
        """
        self.run('ipython')

    def python(self):
        """Runs a python shell.

        Falls back to Python shell if unavailable
        """
        self.run('python')

    @args('--shell',
          metavar='<bpython|ipython|python>',
          help='Python shell')
    def run(self, shell=None):
        """Runs a Python interactive interpreter."""
        if not shell:
            shell = 'bpython'

        if shell == 'bpython':
            try:
                import bpython
                bpython.embed()
            except ImportError:
                shell = 'ipython'
        if shell == 'ipython':
            try:
                from IPython import embed
                embed()
            except ImportError:
                try:
                    # Ipython < 0.11
                    # Explicitly pass an empty list as arguments, because
                    # otherwise IPython would use sys.argv from this script.
                    import IPython

                    shell = IPython.Shell.IPShell(argv=[])
                    shell.mainloop()
                except ImportError:
                    # no IPython module
                    shell = 'python'

        if shell == 'python':
            import code
            try:
                # Try activating rlcompleter, because it's handy.
                import readline
            except ImportError:
                pass
            else:
                # We don't have to wrap the following import in a 'try',
                # because we already know 'readline' was imported successfully.
                import rlcompleter    # noqa
                readline.parse_and_bind("tab:complete")
            code.interact()

    @args('--path', required=True, help='Script path')
    def script(self, path):
        """Runs the script from the specified path with flags set properly."""
        exec(compile(open(path).read(), path, 'exec'), locals(), globals())


def _db_error(caught_exception):
    print('%s' % caught_exception)
    print(_("The above error may show that the database has not "
            "been created.\nPlease create a database using "
            "'cinder-manage db sync' before running this command."))
    sys.exit(1)


class HostCommands(object):
    """List hosts."""

    @args('zone', nargs='?', default=None,
          help='Availability Zone (default: %(default)s)')
    def list(self, zone=None):
        """Show a list of all physical hosts.

        Can be filtered by zone.
        args: [zone]
        """
        print(_("%(host)-25s\t%(zone)-15s") % {'host': 'host', 'zone': 'zone'})
        ctxt = context.get_admin_context()
        services = objects.ServiceList.get_all(ctxt)
        if zone:
            services = [s for s in services if s.availability_zone == zone]
        hosts = []
        for srv in services:
            if not [h for h in hosts if h['host'] == srv['host']]:
                hosts.append(srv)

        for h in hosts:
            print(_("%(host)-25s\t%(availability_zone)-15s")
                  % {'host': h['host'],
                     'availability_zone': h['availability_zone']})


class DbCommands(object):
    """Class for managing the database."""

    # NOTE: Online migrations cannot depend on having Cinder services running.
    # Migrations can be called during Fast-Forward Upgrades without having any
    # Cinder services up.
    # NOTE; Online migrations must be removed at the beginning of the next
    # release to the one they've been introduced.  A comment with the release
    # a migration is introduced and the one where it must be removed must
    # preceed any element of the "online_migrations" tuple, like this:
    #    # Added in Queens remove in Rocky
    #    db.service_uuids_online_data_migration,
    online_migrations = (
        # Added in Train
        db.untyped_volumes_online_data_migration,
        # Added in Train
        db.untyped_snapshots_online_data_migration
    )

    def __init__(self):
        pass

    @args('version', nargs='?', default=None, type=int,
          help='Database version')
    @args('--bump-versions', dest='bump_versions', default=False,
          action='store_true',
          help='Update RPC and Objects versions when doing offline upgrades, '
               'with this we no longer need to restart the services twice '
               'after the upgrade to prevent ServiceTooOld exceptions.')
    def sync(self, version=None, bump_versions=False):
        """Sync the database up to the most recent version."""
        if version is not None and version > db.MAX_INT:
            print(_('Version should be less than or equal to '
                    '%(max_version)d.') % {'max_version': db.MAX_INT})
            sys.exit(1)
        try:
            result = db_migration.db_sync(version)
        except db_exc.DBMigrationError as ex:
            print("Error during database migration: %s" % ex)
            sys.exit(1)

        try:
            if bump_versions:
                ctxt = context.get_admin_context()
                services = objects.ServiceList.get_all(ctxt)
                for service in services:
                    rpc_version = RPC_VERSIONS[service.binary]
                    if (service.rpc_current_version != rpc_version or
                            service.object_current_version != OVO_VERSION):
                        service.rpc_current_version = rpc_version
                        service.object_current_version = OVO_VERSION
                        service.save()
        except Exception as ex:
            print(_('Error during service version bump: %s') % ex)
            sys.exit(2)

        return result

    def version(self):
        """Print the current database version."""
        print(migration.db_version(db_api.get_engine(),
                                   db_migration.MIGRATE_REPO_PATH,
                                   db_migration.INIT_VERSION))

    @args('age_in_days', type=int,
          help='Purge deleted rows older than age in days')
    def purge(self, age_in_days):
        """Purge deleted rows older than a given age from cinder tables."""
        age_in_days = int(age_in_days)
        if age_in_days < 0:
            print(_("Must supply a positive value for age"))
            sys.exit(1)
        if age_in_days >= (int(time.time()) / 86400):
            print(_("Maximum age is count of days since epoch."))
            sys.exit(1)
        ctxt = context.get_admin_context()

        try:
            db.purge_deleted_rows(ctxt, age_in_days)
        except db_exc.DBReferenceError:
            print(_("Purge command failed, check cinder-manage "
                    "logs for more details."))
            sys.exit(1)

    def _run_migration(self, ctxt, max_count):
        ran = 0
        exceptions = False
        migrations = {}
        for migration_meth in self.online_migrations:
            count = max_count - ran
            try:
                found, done = migration_meth(ctxt, count)
            except Exception:
                msg = (_("Error attempting to run %(method)s") %
                       {'method': migration_meth.__name__})
                print(msg)
                LOG.exception(msg)
                exceptions = True
                found = done = 0

            name = migration_meth.__name__
            if found:
                print(_('%(found)i rows matched query %(meth)s, %(done)i '
                        'migrated') % {'found': found,
                                       'meth': name,
                                       'done': done})
            migrations[name] = found, done
            if max_count is not None:
                ran += done
                if ran >= max_count:
                    break
        return migrations, exceptions

    @args('--max_count', metavar='<number>', dest='max_count', type=int,
          help='Maximum number of objects to consider.')
    def online_data_migrations(self, max_count=None):
        """Perform online data migrations for the release in batches."""
        ctxt = context.get_admin_context()
        if max_count is not None:
            unlimited = False
            if max_count < 1:
                print(_('Must supply a positive value for max_count.'))
                sys.exit(127)
        else:
            unlimited = True
            max_count = 50
            print(_('Running batches of %i until complete.') % max_count)

        ran = None
        exceptions = False
        migration_info = {}
        while ran is None or ran != 0:
            migrations, exceptions = self._run_migration(ctxt, max_count)
            ran = 0
            for name in migrations:
                migration_info.setdefault(name, (0, 0))
                migration_info[name] = (
                    max(migration_info[name][0], migrations[name][0]),
                    migration_info[name][1] + migrations[name][1],
                )
                ran += migrations[name][1]
            if not unlimited:
                break
        headers = ["{}".format(_('Migration')),
                   "{}".format(_('Total Needed')),
                   "{}".format(_('Completed')), ]
        t = prettytable.PrettyTable(headers)
        for name in sorted(migration_info.keys()):
            info = migration_info[name]
            t.add_row([name, info[0], info[1]])
        print(t)

        # NOTE(imacdonn): In the "unlimited" case, the loop above will only
        # terminate when all possible migrations have been effected. If we're
        # still getting exceptions, there's a problem that requires
        # intervention. In the max-count case, exceptions are only considered
        # fatal if no work was done by any other migrations ("not ran"),
        # because otherwise work may still remain to be done, and that work
        # may resolve dependencies for the failing migrations.
        if exceptions and (unlimited or not ran):
            print(_("Some migrations failed unexpectedly. Check log for "
                    "details."))
            sys.exit(2)

        sys.exit(1 if ran else 0)

    @args('--enable-replication', action='store_true', default=False,
          help='Set replication status to enabled (default: %(default)s).')
    @args('--active-backend-id', default=None,
          help='Change the active backend ID (default: %(default)s).')
    @args('--backend-host', required=True,
          help='The backend host name.')
    def reset_active_backend(self, enable_replication, active_backend_id,
                             backend_host):
        """Reset the active backend for a host."""

        ctxt = context.get_admin_context()

        try:
            db.reset_active_backend(ctxt, enable_replication,
                                    active_backend_id, backend_host)
        except db_exc.DBReferenceError:
            print(_("Failed to reset active backend for host %s, "
                    "check cinder-manage logs for more details.") %
                  backend_host)
            sys.exit(1)


class VersionCommands(object):
    """Class for exposing the codebase version."""

    def __init__(self):
        pass

    def list(self):
        print(version.version_string())

    def __call__(self):
        self.list()


class VolumeCommands(object):
    """Methods for dealing with a cloud in an odd state."""

    @args('volume_id',
          help='Volume ID to be deleted')
    def delete(self, volume_id):
        """Delete a volume, bypassing the check that it must be available."""
        ctxt = context.get_admin_context()
        volume = objects.Volume.get_by_id(ctxt, volume_id)
        host = volume_utils.extract_host(volume.host) if volume.host else None

        if not host:
            print(_("Volume not yet assigned to host."))
            print(_("Deleting volume from database and skipping rpc."))
            volume.destroy()
            return

        if volume.status == 'in-use':
            print(_("Volume is in-use."))
            print(_("Detach volume from instance and then try again."))
            return

        rpc.init(CONF)
        rpcapi = volume_rpcapi.VolumeAPI()
        rpcapi.delete_volume(ctxt, volume)

    @args('--currenthost', required=True, help='Existing volume host name')
    @args('--newhost', required=True, help='New volume host name')
    def update_host(self, currenthost, newhost):
        """Modify the host name associated with a volume.

        Particularly to recover from cases where one has moved
        their Cinder Volume node, or modified their backend_name in a
        multi-backend config.
        """
        ctxt = context.get_admin_context()
        volumes = db.volume_get_all_by_host(ctxt,
                                            currenthost)
        for v in volumes:
            db.volume_update(ctxt, v['id'],
                             {'host': newhost})


class ConfigCommands(object):
    """Class for exposing the flags defined by flag_file(s)."""

    def __init__(self):
        pass

    @args('param', nargs='?', default=None,
          help='Configuration parameter to display (default: %(default)s)')
    def list(self, param=None):
        """List parameters configured for cinder.

        Lists all parameters configured for cinder unless an optional argument
        is specified.  If the parameter is specified we only print the
        requested parameter.  If the parameter is not found an appropriate
        error is produced by .get*().
        """
        param = param and param.strip()
        if param:
            print('%s = %s' % (param, CONF.get(param)))
        else:
            for key, value in CONF.items():
                print('%s = %s' % (key, value))


class BackupCommands(object):
    """Methods for managing backups."""

    def list(self):
        """List all backups.

        List all backups (including ones in progress) and the host
        on which the backup operation is running.
        """
        ctxt = context.get_admin_context()
        backups = objects.BackupList.get_all(ctxt)

        hdr = "%-32s\t%-32s\t%-32s\t%-24s\t%-24s\t%-12s\t%-12s\t%-12s\t%-12s"
        print(hdr % (_('ID'),
                     _('User ID'),
                     _('Project ID'),
                     _('Host'),
                     _('Name'),
                     _('Container'),
                     _('Status'),
                     _('Size'),
                     _('Object Count')))

        res = "%-32s\t%-32s\t%-32s\t%-24s\t%-24s\t%-12s\t%-12s\t%-12d\t%-12d"
        for backup in backups:
            object_count = 0
            if backup['object_count'] is not None:
                object_count = backup['object_count']
            print(res % (backup['id'],
                         backup['user_id'],
                         backup['project_id'],
                         backup['host'],
                         backup['display_name'],
                         backup['container'],
                         backup['status'],
                         backup['size'],
                         object_count))

    @args('--currenthost', required=True, help='Existing backup host name')
    @args('--newhost', required=True, help='New backup host name')
    def update_backup_host(self, currenthost, newhost):
        """Modify the host name associated with a backup.

        Particularly to recover from cases where one has moved
        their Cinder Backup node, and not set backup_use_same_backend.
        """
        ctxt = context.get_admin_context()
        backups = objects.BackupList.get_all_by_host(ctxt, currenthost)
        for bk in backups:
            bk.host = newhost
            bk.save()


class BaseCommand(object):
    @staticmethod
    def _normalize_time(time_field):
        return time_field and timeutils.normalize_time(time_field)

    @staticmethod
    def _state_repr(is_up):
        return ':-)' if is_up else 'XXX'


class ServiceCommands(BaseCommand):
    """Methods for managing services."""
    def list(self):
        """Show a list of all cinder services."""
        ctxt = context.get_admin_context()
        services = objects.ServiceList.get_all(ctxt)
        print_format = "%-16s %-36s %-16s %-10s %-5s %-20s %-12s %-15s %-36s"
        print(print_format % (_('Binary'),
                              _('Host'),
                              _('Zone'),
                              _('Status'),
                              _('State'),
                              _('Updated At'),
                              _('RPC Version'),
                              _('Object Version'),
                              _('Cluster')))
        for svc in services:
            art = self._state_repr(svc.is_up)
            status = 'disabled' if svc.disabled else 'enabled'
            updated_at = self._normalize_time(svc.updated_at)
            rpc_version = svc.rpc_current_version
            object_version = svc.object_current_version
            cluster = svc.cluster_name or ''
            print(print_format % (svc.binary, svc.host,
                                  svc.availability_zone, status, art,
                                  updated_at, rpc_version, object_version,
                                  cluster))

    @args('binary', type=str,
          help='Service to delete from the host.')
    @args('host_name', type=str,
          help='Host from which to remove the service.')
    def remove(self, binary, host_name):
        """Completely removes a service."""
        ctxt = context.get_admin_context()
        try:
            svc = objects.Service.get_by_args(ctxt, host_name, binary)
            svc.destroy()
        except exception.ServiceNotFound as e:
            print(_("Host not found. Failed to remove %(service)s"
                    " on %(host)s.") %
                  {'service': binary, 'host': host_name})
            print(u"%s" % e.args)
            return 2
        print(_("Service %(service)s on host %(host)s removed.") %
              {'service': binary, 'host': host_name})


class ClusterCommands(BaseCommand):
    """Methods for managing clusters."""
    def list(self):
        """Show a list of all cinder services."""
        ctxt = context.get_admin_context()
        clusters = objects.ClusterList.get_all(ctxt, services_summary=True)
        print_format = "%-36s %-16s %-10s %-5s %-20s %-7s %-12s %-20s"
        print(print_format % (_('Name'),
                              _('Binary'),
                              _('Status'),
                              _('State'),
                              _('Heartbeat'),
                              _('Hosts'),
                              _('Down Hosts'),
                              _('Updated At')))
        for cluster in clusters:
            art = self._state_repr(cluster.is_up)
            status = 'disabled' if cluster.disabled else 'enabled'
            heartbeat = self._normalize_time(cluster.last_heartbeat)
            updated_at = self._normalize_time(cluster.updated_at)
            print(print_format % (cluster.name, cluster.binary, status, art,
                                  heartbeat, cluster.num_hosts,
                                  cluster.num_down_hosts, updated_at))

    @args('--recursive', action='store_true', default=False,
          help='Delete associated hosts.')
    @args('binary', type=str,
          help='Service to delete from the cluster.')
    @args('cluster-name', type=str, help='Cluster to delete.')
    def remove(self, recursive, binary, cluster_name):
        """Completely removes a cluster."""
        ctxt = context.get_admin_context()
        try:
            cluster = objects.Cluster.get_by_id(ctxt, None, name=cluster_name,
                                                binary=binary,
                                                get_services=recursive)
        except exception.ClusterNotFound:
            print(_("Couldn't remove cluster %s because it doesn't exist.") %
                  cluster_name)
            return 2

        if recursive:
            for service in cluster.services:
                service.destroy()

        try:
            cluster.destroy()
        except exception.ClusterHasHosts:
            print(_("Couldn't remove cluster %s because it still has hosts.") %
                  cluster_name)
            return 2

        msg = _('Cluster %s successfully removed.') % cluster_name
        if recursive:
            msg = (_('%(msg)s And %(num)s services from the cluster were also '
                     'removed.') % {'msg': msg, 'num': len(cluster.services)})
        print(msg)

    @args('--full-rename', dest='partial',
          action='store_false', default=True,
          help='Do full cluster rename instead of just replacing provided '
               'current cluster name and preserving backend and/or pool info.')
    @args('current', help='Current cluster name.')
    @args('new', help='New cluster name.')
    def rename(self, partial, current, new):
        """Rename cluster name for Volumes and Consistency Groups.

        Useful when you want to rename a cluster, particularly when the
        backend_name has been modified in a multi-backend config or we have
        moved from a single backend to multi-backend.
        """
        ctxt = context.get_admin_context()

        # Convert empty strings to None
        current = current or None
        new = new or None

        # Update Volumes
        num_vols = objects.VolumeList.include_in_cluster(
            ctxt, new, partial_rename=partial, cluster_name=current)

        # Update Consistency Groups
        num_cgs = objects.ConsistencyGroupList.include_in_cluster(
            ctxt, new, partial_rename=partial, cluster_name=current)

        if num_vols or num_cgs:
            msg = _('Successfully renamed %(num_vols)s volumes and '
                    '%(num_cgs)s consistency groups from cluster %(current)s '
                    'to %(new)s')
            print(msg % {'num_vols': num_vols, 'num_cgs': num_cgs, 'new': new,
                         'current': current})
        else:
            msg = _('No volumes or consistency groups exist in cluster '
                    '%(current)s.')
            print(msg % {'current': current})
            return 2


class ConsistencyGroupCommands(object):
    """Methods for managing consistency groups."""

    @args('--currenthost', required=True, help='Existing CG host name')
    @args('--newhost', required=True, help='New CG host name')
    def update_cg_host(self, currenthost, newhost):
        """Modify the host name associated with a Consistency Group.

        Particularly to recover from cases where one has moved
        a host from single backend to multi-backend, or changed the host
        configuration option, or modified the backend_name in a multi-backend
        config.
        """

        ctxt = context.get_admin_context()
        groups = objects.ConsistencyGroupList.get_all(
            ctxt, {'host': currenthost})
        for gr in groups:
            gr.host = newhost
            gr.save()


CATEGORIES = {
    'backup': BackupCommands,
    'config': ConfigCommands,
    'cluster': ClusterCommands,
    'cg': ConsistencyGroupCommands,
    'db': DbCommands,
    'host': HostCommands,
    'service': ServiceCommands,
    'shell': ShellCommands,
    'version': VersionCommands,
    'volume': VolumeCommands,
}


def methods_of(obj):
    """Return non-private methods from an object.

    Get all callable methods of an object that don't start with underscore
    :return: a list of tuples of the form (method_name, method)
    """
    result = []
    for i in dir(obj):
        if isinstance(getattr(obj, i),
                      collections.Callable) and not i.startswith('_'):
            result.append((i, getattr(obj, i)))
    return result


def add_command_parsers(subparsers):
    for category in sorted(CATEGORIES):
        command_object = CATEGORIES[category]()

        parser = subparsers.add_parser(category)
        parser.set_defaults(command_object=command_object)

        category_subparsers = parser.add_subparsers(dest='action')

        for (action, action_fn) in methods_of(command_object):
            parser = category_subparsers.add_parser(action)

            action_kwargs = []
            for args, kwargs in getattr(action_fn, 'args', []):
                parser.add_argument(*args, **kwargs)

            parser.set_defaults(action_fn=action_fn)
            parser.set_defaults(action_kwargs=action_kwargs)


category_opt = cfg.SubCommandOpt('category',
                                 title='Command categories',
                                 handler=add_command_parsers)


def get_arg_string(args):
    if args[0] == '-':
        # (Note)zhiteng: args starts with FLAGS.oparser.prefix_chars
        # is optional args. Notice that cfg module takes care of
        # actual ArgParser so prefix_chars is always '-'.
        if args[1] == '-':
            # This is long optional arg
            args = args[2:]
        else:
            args = args[1:]

    # We convert dashes to underscores so we can have cleaner optional arg
    # names
    if args:
        args = args.replace('-', '_')

    return args


def fetch_func_args(func):
    fn_kwargs = {}
    for args, kwargs in getattr(func, 'args', []):
        # Argparser `dest` configuration option takes precedence for the name
        arg = kwargs.get('dest') or get_arg_string(args[0])
        fn_kwargs[arg] = getattr(CONF.category, arg)

    return fn_kwargs


def main():
    objects.register_all()
    """Parse options and call the appropriate class/method."""
    CONF.register_cli_opt(category_opt)
    script_name = sys.argv[0]
    if len(sys.argv) < 2:
        print(_("\nOpenStack Cinder version: %(version)s\n") %
              {'version': version.version_string()})
        print(script_name + " category action [<args>]")
        print(_("Available categories:"))
        for category in CATEGORIES:
            print(_("\t%s") % category)
        sys.exit(2)

    try:
        CONF(sys.argv[1:], project='cinder',
             version=version.version_string())
        logging.setup(CONF, "cinder")
        python_logging.captureWarnings(True)
    except cfg.ConfigDirNotFoundError as details:
        print(_("Invalid directory: %s") % details)
        sys.exit(2)
    except cfg.ConfigFilesNotFoundError as e:
        cfg_files = e.config_files
        print(_("Failed to read configuration file(s): %s") % cfg_files)
        sys.exit(2)

    fn = CONF.category.action_fn
    fn_kwargs = fetch_func_args(fn)
    fn(**fn_kwargs)



# Copyright (c) 2018 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import json

import re
import requests


class FakeBaseSession(requests.Session):
    method_map = {}

    def _get_response(self, method, url):
        url_map = self.method_map.get(method, {})
        tmp = None
        data = {}
        for k in url_map:
            if re.search(k, url):
                if not tmp or len(tmp) < len(k):
                    data = url_map[k]
                    tmp = k

        resp_content = {'result': 0}
        resp_content.update(data)
        resp = requests.Response()
        resp.headers['X-Auth-Token'] = 'fake_token'
        resp.status_code = 0
        resp.encoding = 'utf-8'
        resp._content = json.dumps(resp_content).encode('utf-8')

        return resp

    def get(self, url, **kwargs):
        return self._get_response('get', url)

    def post(self, url, **kwargs):
        return self._get_response('post', url)
